# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0
"""Find Tesseract and Ghostscript binaries on Windows using the registry."""

from __future__ import annotations

import logging
import os
import re
import shutil
import sys
from collections.abc import Callable, Iterable, Iterator
from itertools import chain
from pathlib import Path
from typing import Any, TypeAlias, TypeVar

from packaging.version import InvalidVersion, Version

if sys.platform == 'win32':
    # mypy understands 'if sys.platform' better than try/except ModuleNotFoundError
    import winreg  # pylint: disable=import-error

    HKEYType: TypeAlias = winreg.HKEYType
else:
    from unittest.mock import Mock

    winreg = Mock(
        spec=['HKEYType', 'EnumKey', 'EnumValue', 'HKEY_LOCAL_MACHINE', 'OpenKey']
    )
    # mypy does not understand winreg.HKeyType where winreg is a Mock (fair enough!)
    HKEYType: TypeAlias = Any  # type: ignore


log = logging.getLogger(__name__)

T = TypeVar('T')
Tkey = TypeVar('Tkey')


def ghostscript_version_key(s: str) -> tuple[int, int, int]:
    """Compare Ghostscript version numbers."""
    try:
        release = [int(elem) for elem in s.split('.', maxsplit=3)]
        while len(release) < 3:
            release.append(0)
        return (release[0], release[1], release[2])
    except ValueError:
        return (0, 0, 0)


def registry_enum(key: HKEYType, enum_fn: Callable[[HKEYType, int], T]) -> Iterator[T]:
    limit = 999
    n = 0
    while n < limit:
        try:
            yield enum_fn(key, n)
            n += 1
        except OSError:
            break
    if n == limit:
        raise ValueError(f"Too many registry keys under {key}")


def registry_subkeys(key: HKEYType) -> Iterator[str]:
    return registry_enum(key, winreg.EnumKey)


def registry_values(key: HKEYType) -> Iterator[tuple[str, Any, int]]:
    return registry_enum(key, winreg.EnumValue)


def registry_path_ghostscript(env=None) -> Iterator[Path]:
    del env  # unused (but needed for protocol)
    try:
        with winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Artifex\GPL Ghostscript"
        ) as k:
            latest_gs = max(
                registry_subkeys(k), key=ghostscript_version_key, default=(0, 0, 0)
            )
        with winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE, rf"SOFTWARE\Artifex\GPL Ghostscript\{latest_gs}"
        ) as k:
            for _, gs_path, _ in registry_values(k):
                yield Path(gs_path) / 'bin'
    except OSError as e:
        log.warning(e)


def registry_path_tesseract(env=None) -> Iterator[Path]:
    del env  # unused (but needed for protocol)
    try:
        with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Tesseract-OCR") as k:
            for subkey, val, _valtype in registry_values(k):
                if subkey == 'InstallDir':
                    tesseract_path = Path(val)
                    yield tesseract_path
    except OSError as e:
        log.warning(e)


def _gs_version_in_path_key(path: Path) -> tuple[str, Version | None]:
    """Key function for comparing Ghostscript and Tesseract paths.

    Ghostscript installs on Windows:
        %PROGRAMFILES%/gs/gs9.56.1/bin -> ('gs', Version('9.56.1'))
        %PROGRAMFILES%/gs/9.24/bin -> ('gs', Version('9.24'))

    Tesseract looks like:
        %PROGRAMFILES%/Tesseract-OCR -> ('Tesseract-OCR', None)

    Thus ensuring the resulting tuple will order the alternatives correctly,
    e.g. gs10.0 > gs9.99.
    """
    match = re.search(r'gs[/\\]?([0-9.]+)[/\\]bin', str(path))
    if match:
        try:
            version_str = match.group(1)
            version = Version(version_str)
            return 'gs', version
        except InvalidVersion:
            pass
    return path.name, None


def program_files_paths(env=None) -> Iterator[Path]:
    if not env:
        env = os.environ
    program_files = env.get('PROGRAMFILES', '')

    def path_walker() -> Iterator[Path]:
        for path in Path(program_files).iterdir():
            if not path.is_dir():
                continue
            if path.name.lower() == 'tesseract-ocr':
                yield path
            elif path.name.lower() == 'gs':
                yield from (p for p in path.glob('**/bin') if p.is_dir())

    return iter(
        sorted(
            (p for p in path_walker()),
            key=_gs_version_in_path_key,
            reverse=True,
        )
    )


def paths_from_env(env=None) -> Iterator[Path]:
    return (Path(p) for p in os.get_exec_path(env) if p)


def shim_path(new_paths: Callable[[Any], Iterator[Path]], env=None) -> str:
    if not env:
        env = os.environ
    return os.pathsep.join(str(p) for p in new_paths(env) if p)


SHIMS = [
    paths_from_env,
    registry_path_ghostscript,
    registry_path_tesseract,
    program_files_paths,
]


def fix_windows_args(program: str, args, env):
    """Adjust our desired program and command line arguments for use on Windows."""
    # If we are running a .py on Windows, ensure we call it with this Python
    # (to support test suite shims)
    if program.lower().endswith('.py'):
        args = [sys.executable] + args

    # If the program we want is not on the PATH, check elsewhere
    for shim in SHIMS:
        shimmed_path = shim_path(shim, env)
        new_args0 = shutil.which(args[0], path=shimmed_path)
        if new_args0:
            args[0] = new_args0
            break

    return args


def unique_everseen(iterable: Iterable[T], key: Callable[[T], Tkey]) -> Iterator[T]:
    """List unique elements, preserving order."""
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    seen: set[Tkey] = set()
    seen_add = seen.add
    for element in iterable:
        k = key(element)
        if k not in seen:
            seen_add(k)
            yield element


def _casefold_path(path: Path) -> str:
    return str.casefold(str(path))


def shim_env_path(env=None):
    if env is None:
        env = os.environ

    shim_paths = chain.from_iterable(shim(env) for shim in SHIMS)
    return os.pathsep.join(
        str(p) for p in unique_everseen(shim_paths, key=_casefold_path)
    )



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""Interface to Tesseract executable."""

from __future__ import annotations

import logging
import re
from contextlib import suppress
from math import pi
from os import fspath
from pathlib import Path
from subprocess import PIPE, STDOUT, CalledProcessError, TimeoutExpired

from packaging.version import Version

from ocrmypdf.exceptions import (
    MissingDependencyError,
    SubprocessOutputError,
    TesseractConfigError,
)
from ocrmypdf.pluginspec import OrientationConfidence
from ocrmypdf.subprocess import get_version, run

log = logging.getLogger(__name__)


TESSERACT_THRESHOLDING_METHODS: dict[str, int] = {
    'auto': 0,
    'otsu': 0,
    'adaptive-otsu': 1,
    'sauvola': 2,
}


class TesseractLoggerAdapter(logging.LoggerAdapter):
    """Prepend [tesseract] to messages emitted from tesseract."""

    def process(self, msg, kwargs):
        kwargs['extra'] = self.extra
        return f'[tesseract] {msg}', kwargs


TESSERACT_VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
        (?P<date>
            [-_\.]
            (?:20[0-9][0-9] [0-1][0-9] [0-3][0-9])       # yyyy mm dd
        )?
        (?P<gitcount>
            [-_\.]?
            [0-9]+
        )?
        (?P<gitcommit>
            [-_\.]?
            g[0-9a-f]{2,10}
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""


class TesseractVersion(Version):
    """Modify standard packaging.Version regex to support Tesseract idiosyncrasies."""

    _regex = re.compile(
        r"^\s*" + TESSERACT_VERSION_PATTERN + r"\s*$", re.VERBOSE | re.IGNORECASE
    )


def version() -> Version:
    return TesseractVersion(get_version('tesseract', regex=r'tesseract\s(.+)'))


def has_thresholding() -> bool:
    """Does Tesseract have -c thresholding method capability?"""
    return version() >= Version('5.0')


def get_languages() -> set[str]:
    def lang_error(output):
        msg = (
            "Tesseract failed to report available languages.\n"
            "Output from Tesseract:\n"
            "-----------\n"
        )
        msg += output
        return msg

    args_tess = ['tesseract', '--list-langs']
    try:
        proc = run(
            args_tess,
            text=True,
            stdout=PIPE,
            stderr=STDOUT,
            logs_errors_to_stdout=True,
            check=True,
        )
        output = proc.stdout
    except CalledProcessError as e:
        raise MissingDependencyError(lang_error(e.output)) from e

    for line in output.splitlines():
        if line.startswith('Error'):
            raise MissingDependencyError(lang_error(output))
    _header, *rest = output.splitlines()
    return {lang.strip() for lang in rest}


def tess_base_args(langs: list[str], engine_mode: int | None) -> list[str]:
    args = ['tesseract']
    if langs:
        args.extend(['-l', '+'.join(langs)])
    if engine_mode is not None:
        args.extend(['--oem', str(engine_mode)])
    return args


def _parse_tesseract_output(binary_output: bytes) -> dict[str, str]:
    def gen():
        for line in binary_output.decode().splitlines():
            line = line.strip()
            parts = line.split(':', maxsplit=2)
            if len(parts) == 2:
                yield parts[0].strip(), parts[1].strip()

    return dict(gen())


def get_orientation(
    input_file: Path, engine_mode: int | None, timeout: float
) -> OrientationConfidence:
    args_tesseract = tess_base_args(['osd'], engine_mode) + [
        '--psm',
        '0',
        fspath(input_file),
        'stdout',
    ]

    try:
        p = run(args_tesseract, stdout=PIPE, stderr=STDOUT, timeout=timeout, check=True)
    except TimeoutExpired:
        return OrientationConfidence(angle=0, confidence=0.0)
    except CalledProcessError as e:
        tesseract_log_output(e.stdout)
        tesseract_log_output(e.stderr)
        if (
            b'Too few characters. Skipping this page' in e.output
            or b'Image too large' in e.output
        ):
            return OrientationConfidence(0, 0)
        raise SubprocessOutputError() from e

    osd = _parse_tesseract_output(p.stdout)
    angle = int(osd.get('Orientation in degrees', 0))
    orient_conf = OrientationConfidence(
        angle=angle, confidence=float(osd.get('Orientation confidence', 0))
    )
    return orient_conf


def get_deskew(
    input_file: Path, languages: list[str], engine_mode: int | None, timeout: float
) -> float:
    """Gets angle to deskew this page, in degrees."""
    args_tesseract = tess_base_args(languages, engine_mode) + [
        '--psm',
        '2',
        fspath(input_file),
        'stdout',
    ]

    try:
        p = run(args_tesseract, stdout=PIPE, stderr=STDOUT, timeout=timeout, check=True)
    except TimeoutExpired:
        return 0.0
    except CalledProcessError as e:
        tesseract_log_output(e.stdout)
        tesseract_log_output(e.stderr)
        if b'Empty page!!' in e.output or (
            e.output == b'' and e.returncode == 1
        ):  # Not enough info for a skew angle - Tess 4 and 5 return different errors
            return 0.0

        raise SubprocessOutputError() from e

    parsed = _parse_tesseract_output(p.stdout)
    deskew_radians = float(parsed.get('Deskew angle', 0))
    deskew_degrees = 180 / pi * deskew_radians
    log.debug(f"Deskew angle: {deskew_degrees:.3f}")
    return deskew_degrees


def tesseract_log_output(stream: bytes) -> None:
    tlog = TesseractLoggerAdapter(
        log,
        extra=log.extra if hasattr(log, 'extra') else None,  # type: ignore
    )

    if not stream:
        return
    try:
        text = stream.decode()
    except UnicodeDecodeError:
        text = stream.decode('utf-8', 'ignore')

    lines = text.splitlines()
    for line in lines:
        if line.startswith("Tesseract Open Source"):
            continue
        elif line.startswith("Warning in pixReadMem"):
            continue
        elif 'diacritics' in line:
            tlog.warning("lots of diacritics - possibly poor OCR")
        elif line.startswith('OSD: Weak margin'):
            tlog.warning("unsure about page orientation")
        elif 'Error in pixScanForForeground' in line:
            pass  # Appears to be spurious/problem with nonwhite borders
        elif 'Error in boxClipToRectangle' in line:
            pass  # Always appears with pixScanForForeground message
        elif 'parameter not found: ' in line.lower():
            tlog.error(line.strip())
            problem = line.split('found: ')[1]
            raise TesseractConfigError(problem)
        elif 'error' in line.lower() or 'exception' in line.lower():
            tlog.error(line.strip())
        elif 'warning' in line.lower():
            tlog.warning(line.strip())
        elif 'read_params_file' in line.lower():
            tlog.error(line.strip())
        else:
            tlog.info(line.strip())


def page_timedout(timeout: float) -> None:
    if timeout == 0:
        return
    log.warning("[tesseract] took too long to OCR - skipping")


def _generate_null_hocr(output_hocr: Path, output_text: Path, image: Path) -> None:
    """Produce an empty .hocr file.

    Ensures page is the same size as the input image.
    """
    output_hocr.write_text('', encoding='utf-8')
    output_text.write_text('[skipped page]', encoding='utf-8')


def generate_hocr(
    *,
    input_file: Path,
    output_hocr: Path,
    output_text: Path,
    languages: list[str],
    engine_mode: int,
    tessconfig: list[str],
    timeout: float,
    pagesegmode: int,
    thresholding: int,
    user_words,
    user_patterns,
) -> None:
    """Generate a hOCR file, which must be converted to PDF."""
    prefix = output_hocr.with_suffix('')

    args_tesseract = tess_base_args(languages, engine_mode)

    if pagesegmode is not None:
        args_tesseract.extend(['--psm', str(pagesegmode)])

    if thresholding != 0 and has_thresholding():
        args_tesseract.extend(['-c', f'thresholding_method={thresholding}'])

    if user_words:
        args_tesseract.extend(['--user-words', user_words])

    if user_patterns:
        args_tesseract.extend(['--user-patterns', user_patterns])

    # Reminder: test suite tesseract test plugins will break after any changes
    # to the number of order parameters here
    args_tesseract.extend([fspath(input_file), fspath(prefix), 'hocr', 'txt'])
    args_tesseract.extend(tessconfig)
    try:
        p = run(args_tesseract, stdout=PIPE, stderr=STDOUT, timeout=timeout, check=True)
        stdout = p.stdout
    except TimeoutExpired:
        # Generate a HOCR file with no recognized text if tesseract times out
        # Temporary workaround to hocrTransform not being able to function if
        # it does not have a valid hOCR file.
        page_timedout(timeout)
        _generate_null_hocr(output_hocr, output_text, input_file)
    except CalledProcessError as e:
        tesseract_log_output(e.output)
        if b'Image too large' in e.output or b'Empty page!!' in e.output:
            _generate_null_hocr(output_hocr, output_text, input_file)
            return

        raise SubprocessOutputError() from e
    else:
        tesseract_log_output(stdout)
        # The sidecar text file will get the suffix .txt; rename it to
        # whatever caller wants it named
        with suppress(FileNotFoundError):
            prefix.with_suffix('.txt').replace(output_text)


def use_skip_page(output_pdf: Path, output_text: Path) -> None:
    output_text.write_text('[skipped page]', encoding='utf-8')

    # A 0 byte file to the output to indicate a skip
    output_pdf.write_bytes(b'')


def generate_pdf(
    *,
    input_file: Path,
    output_pdf: Path,
    output_text: Path,
    languages: list[str],
    engine_mode: int,
    tessconfig: list[str],
    timeout: float,
    pagesegmode: int,
    thresholding: int,
    user_words,
    user_patterns,
) -> None:
    """Generate a PDF using Tesseract's internal PDF generator.

    We specifically a text-only PDF which is more suitable for combining with
    the input page.
    """
    args_tesseract = tess_base_args(languages, engine_mode)

    if pagesegmode is not None:
        args_tesseract.extend(['--psm', str(pagesegmode)])

    args_tesseract.extend(['-c', 'textonly_pdf=1'])

    if thresholding != 0 and has_thresholding():
        args_tesseract.extend(['-c', f'thresholding_method={thresholding}'])

    if user_words:
        args_tesseract.extend(['--user-words', user_words])

    if user_patterns:
        args_tesseract.extend(['--user-patterns', user_patterns])

    prefix = output_pdf.parent / Path(output_pdf.stem)

    # Reminder: test suite tesseract test plugins might break after any changes
    # to the number of order parameters here

    args_tesseract.extend([fspath(input_file), fspath(prefix), 'pdf', 'txt'])
    args_tesseract.extend(tessconfig)
    try:
        p = run(args_tesseract, stdout=PIPE, stderr=STDOUT, timeout=timeout, check=True)
        stdout = p.stdout
        with suppress(FileNotFoundError):
            prefix.with_suffix('.txt').replace(output_text)
    except TimeoutExpired:
        page_timedout(timeout)
        use_skip_page(output_pdf, output_text)
    except CalledProcessError as e:
        tesseract_log_output(e.output)
        if b'Image too large' in e.output or b'Empty page!!' in e.output:
            use_skip_page(output_pdf, output_text)
            return
        raise SubprocessOutputError() from e
    else:
        tesseract_log_output(stdout)



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""Logging support classes."""

from __future__ import annotations

import logging

from rich.console import Console
from rich.logging import RichHandler


class PageNumberFilter(logging.Filter):
    """Insert PDF page number that emitted log message to log record."""

    def filter(self, record):
        pageno = getattr(record, 'pageno', None)
        if isinstance(pageno, int):
            record.pageno = f'{pageno:5d} '
        elif pageno is None:
            record.pageno = ''
        return True


class RichLoggingHandler(RichHandler):
    def __init__(self, console: Console, **kwargs):
        super().__init__(
            console=console, show_level=False, show_time=False, markup=False, **kwargs
        )



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""OCRmyPDF pluggy plugin specification."""

from __future__ import annotations

from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from collections.abc import Sequence, Set
from logging import Handler
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

import pluggy

from ocrmypdf import Executor, PdfContext
from ocrmypdf._progressbar import ProgressBar
from ocrmypdf.helpers import Resolution

if TYPE_CHECKING:
    from PIL import Image

    # pylint: disable=ungrouped-imports
    from ocrmypdf._jobcontext import PageContext
    from ocrmypdf.pdfinfo import PdfInfo

    # pylint: enable=ungrouped-imports

hookspec = pluggy.HookspecMarker('ocrmypdf')

# pylint: disable=unused-argument
# mypy: disable-error-code=empty-body


@hookspec(firstresult=True)
def get_logging_console() -> Handler:
    """Returns a custom logging handler.

    Generally this is necessary when both logging output and a progress bar are both
    outputting to ``sys.stderr``.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


@hookspec
def initialize(plugin_manager: pluggy.PluginManager) -> None:
    """Called when this plugin is first loaded into OCRmyPDF.

    The primary intended use of this is for plugins to check compatibility with other
    plugins and possibly block other blocks, a plugin that wishes to block ocrmypdf's
    built-in optimize plugin could do:

    .. code-block::

        plugin_manager.set_blocked('ocrmypdf.builtin_plugins.optimize')

    It would also be reasonable for an plugin implementation to check if it is unable
    to proceed, for example, because a required dependency is missing. (If the plugin's
    ability to proceed depends on options and arguments, use ``validate`` instead.)

    Raises:
        ocrmypdf.exceptions.ExitCodeException: If options are not acceptable
            and the application should terminate gracefully with an informative
            message and error code.

    Note:
        This hook will be called from the main process, and may modify global state
        before child worker processes are forked.
    """


@hookspec
def add_options(parser: ArgumentParser) -> None:
    """Allows the plugin to add its own command line and API arguments.

    OCRmyPDF converts command line arguments to API arguments, so adding
    arguments here will cause new arguments to be processed for API calls
    to ``ocrmypdf.ocr``, or when invoked on the command line.

    Note:
        This hook will be called from the main process, and may modify global state
        before child worker processes are forked.
    """


@hookspec
def check_options(options: Namespace) -> None:
    """Called to ask the plugin to check all of the options.

    The plugin may check if options that it added are valid.

    Warnings or other messages may be passed to the user by creating a logger
    object using ``log = logging.getLogger(__name__)`` and logging to this.

    The plugin may also modify the *options*. All objects that are in options
    must be picklable so they can be marshalled to child worker processes.

    Raises:
        ocrmypdf.exceptions.ExitCodeException: If options are not acceptable
            and the application should terminate gracefully with an informative
            message and error code.

    Note:
        This hook will be called from the main process, and may modify global state
        before child worker processes are forked.
    """


@hookspec(firstresult=True)
def get_executor(progressbar_class: type[ProgressBar]) -> Executor:
    """Called to obtain an object that manages parallel execution.

    This may be used to replace OCRmyPDF's default parallel execution system
    with a third party alternative. For example, you could make OCRmyPDF run in a
    distributed environment.

    OCRmyPDF's executors are analogous to the standard Python executors in
    ``conconcurrent.futures``, but they do not work the same way. Executors may
    be reused for different, unrelated batch operations, since all of the context
    for a given job are passed to :meth:`Executor.__call__`.

    Should be of type :class:`Executor` or otherwise conforming to the protocol
    of that call.

    Arguments:
        progressbar_class: A progress bar class, which will be created when

    Note:
        This hook will be called from the main process, and may modify global state
        before child worker processes are forked.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


@hookspec(firstresult=True)
def get_progressbar_class() -> type[ProgressBar]:
    """Called to obtain a class that can be used to monitor progress.

    OCRmyPDF will call this function when it wants to display a progress bar.
    The class returned by this function must be compatible with the
    :class:`ProgressBar` protocol.

    Example:
        Here is how OCRmyPDF will use the progress bar:

        .. code-block:: python

            pbar_class = pm.hook.get_progressbar_class()
            with pbar_class(**progress_kwargs) as pbar:
                ... # do some work
                pbar.update(1)
    """


@hookspec
def validate(pdfinfo: PdfInfo, options: Namespace) -> None:
    """Called to give a plugin an opportunity to review *options* and *pdfinfo*.

    *options* contains the "work order" to process a particular file. *pdfinfo*
    contains information about the input file obtained after loading and
    parsing. The plugin may modify the *options*. For example, you could decide
    that a certain type of file should be treated with ``options.force_ocr = True``
    based on information in its *pdfinfo*.

    Raises:
        ocrmypdf.exceptions.ExitCodeException: If options or pdfinfo are not acceptable
            and the application should terminate gracefully with an informative
            message and error code.

    Note:
        This hook will be called from the main process, and may modify global state
        before child worker processes are forked.
    """


@hookspec(firstresult=True)
def rasterize_pdf_page(
    input_file: Path,
    output_file: Path,
    raster_device: str,
    raster_dpi: Resolution,
    pageno: int,
    page_dpi: Resolution | None,
    rotation: int | None,
    filter_vector: bool,
    stop_on_soft_error: bool,
) -> Path:
    """Rasterize one page of a PDF at resolution raster_dpi in canvas units.

    The image is sized to match the integer pixels dimensions implied by
    raster_dpi even if those numbers are noninteger. The image's DPI will
    be overridden with the values in page_dpi.

    Args:
        input_file: The PDF to rasterize.
        output_file: The desired name of the rasterized image.
        raster_device: Type of image to produce at output_file.
        raster_dpi: Resolution in dots per inch at which to rasterize page.
        pageno: Page number to rasterize (beginning at page 1).
        page_dpi: Resolution, overriding output image DPI.
        rotation: Cardinal angle, clockwise, to rotate page.
        filter_vector: If True, remove vector graphics objects.
        stop_on_soft_error: If there is an "soft error" such that PDF page image
            generation can proceed, but may visually differ from the original,
            the implementer of this hook should raise a detailed exception. If
            ``False``, continue processing and report by logging it. If the hook
            cannot proceed, it should always raise an exception, regardless of
            this setting. One "soft error" would be a missing font that is
            required to properly rasterize the PDF.

    Returns:
        Path: output_file if successful

    Note:
        This hook will be called from child processes. Modifying global state
        will not affect the main process or other child processes.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


@hookspec(firstresult=True)
def filter_ocr_image(page: PageContext, image: Image.Image) -> Image.Image:
    """Called to filter the image before it is sent to OCR.

    This is the image that OCR sees, not what the user sees when they view the
    PDF. In certain modes such as ``--redo-ocr``, portions of the image may be
    masked out to hide them from OCR.

    The main uses of this hook are expected to be hiding content from OCR,
    conditioning images to OCR better with filters, and adjusting images to
    match any constraints imposed by the OCR engine.

    The input image may be color, grayscale, or monochrome, and the
    output image may differ. For example, if you know that a custom OCR engine
    does not care about the color of the text, you could convert the image to
    it to grayscale or monochrome.

    Generally speaking, the output image should be a faithful representation of
    of the input image. You *may* change the pixel width and height of the
    the input image, but you must not change the aspect ratio, and you must
    calculate the DPI of the output image based on the new pixel width and
    height or the OCR text layer will be misaligned with the visual position.

    The built-in Tesseract OCR engine uses this hook itself to downsample
    very large images to fit its constraints.

    Note:
        This hook will be called from child processes. Modifying global state
        will not affect the main process or other child processes.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


@hookspec(firstresult=True)
def filter_page_image(page: PageContext, image_filename: Path) -> Path:
    """Called to filter the whole page before it is inserted into the PDF.

    A whole page image is only produced when preprocessing command line arguments
    are issued or when ``--force-ocr`` is issued. If no whole page is image is
    produced for a given page, this function will not be called. This is not
    the image that will be shown to OCR.

    If the function does not want to modify the image, it should return
    ``image_filename``. The hook may overwrite ``image_filename`` with a new file.

    The output image should preserve the same physical unit dimensions, that is
    ``(width * dpi_x, height * dpi_y)``. That is, if the image is resized, the DPI
    must be adjusted by the reciprocal. If this is not preserved, the PDF page
    will be resized and the OCR layer misaligned. OCRmyPDF does nothing
    to enforce these constraints; it is up to the plugin to do sensible things.

    OCRmyPDF will create the PDF page based on the image format used (unless the
    hook is overridden). If you convert the image to a JPEG, the output page will
    be created as a JPEG, etc. If you change the colorspace, that change will be
    kept. Note that the OCRmyPDF image optimization stage, if enabled, may
    ultimately chose a different format.

    If the return value is a file that does not exist, ``FileNotFoundError``
    will occur. The return value should be a path to a file in the same folder
    as ``image_filename``.

    Note:
        This hook will be called from child processes. Modifying global state
        will not affect the main process or other child processes.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


@hookspec(firstresult=True)
def filter_pdf_page(page: PageContext, image_filename: Path, output_pdf: Path) -> Path:
    """Called to convert a filtered whole page image into a PDF.

    A whole page image is only produced when preprocessing command line arguments
    are issued or when ``--force-ocr`` is issued. If no whole page is image is
    produced for a given page, this function will not be called. This is not
    the image that will be shown to OCR. The whole page image is filtered in
    the hook above, ``filter_page_image``, then this function is called for
    PDF conversion.

    This function will only be called when OCRmyPDF runs in a mode such as
    "force OCR" mode where rasterizing of all content is performed.

    Clever things could be done at this stage such as segmenting the page image into
    color regions or vector equivalents.

    The provider of the hook implementation is responsible for ensuring that the
    OCR text layer is aligned with the PDF produced here, or text misalignment
    will result.

    Currently this function must produce a single page PDF or the pipeline will
    fail.  If the intent is to remove the PDF, then create a single page empty
    PDF.

    Args:
        page: Context for this page.
        image_filename: Filename of the input image used to create output_pdf,
            for "reference" if recreating the output_pdf entirely.
        output_pdf: The previous created output_pdf.

    Returns:
        output_pdf

    Note:
        This hook will be called from child processes. Modifying global state
        will not affect the main process or other child processes.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


class OrientationConfidence(NamedTuple):
    """Expresses an OCR engine's confidence in page rotation.

    Attributes:
        angle: The clockwise angle (0, 90, 180, 270) that the page should be
            rotated. 0 means no rotation.
        confidence: How confident the OCR engine is that this the correct
            rotation. 0 is not confident, 15 is very confident. Arbitrary units.
    """

    angle: int
    confidence: float


class OcrEngine(ABC):
    """A class representing an OCR engine with capabilities similar to Tesseract OCR.

    This could be used to create a plugin for another OCR engine instead of
    Tesseract OCR.
    """

    @staticmethod
    @abstractmethod
    def version() -> str:
        """Returns the version of the OCR engine."""

    @staticmethod
    @abstractmethod
    def creator_tag(options: Namespace) -> str:
        """Returns the creator tag to identify this software's role in creating the PDF.

        This tag will be inserted in the XMP metadata and DocumentInfo dictionary
        as appropriate. Ideally you should include the name of the OCR engine and its
        version. The text should not contain line breaks. This is to help developers
        like yourself identify the software that produced this file.

        OCRmyPDF will always prepend its name to this value.
        """

    @abstractmethod
    def __str__(self):
        """Returns name of OCR engine and version.

        This is used when OCRmyPDF wants to mention the name of the OCR engine
        to the user, usually in an error message.
        """

    @staticmethod
    @abstractmethod
    def languages(options: Namespace) -> Set[str]:
        """Returns the set of all languages that are supported by the engine.

        Languages are typically given in 3-letter ISO 3166-1 codes, but actually
        can be any value understood by the OCR engine.
        """

    @staticmethod
    @abstractmethod
    def get_orientation(input_file: Path, options: Namespace) -> OrientationConfidence:
        """Returns the orientation of the image."""

    @staticmethod
    def get_deskew(input_file: Path, options: Namespace) -> float:
        """Returns the deskew angle of the image, in degrees."""
        return 0.0

    @staticmethod
    @abstractmethod
    def generate_hocr(
        input_file: Path, output_hocr: Path, output_text: Path, options: Namespace
    ) -> None:
        """Called to produce a hOCR file from a page image and sidecar text file.

        A hOCR file is an HTML-like file that describes the position of text on a
        page. OCRmyPDF can create a text only PDF from the hOCR file and graft it
        onto the output PDF.

        This function executes in a worker thread or worker process. OCRmyPDF
        automatically parallelizes OCR over pages. The OCR engine should not
        introduce more parallelism.

        Args:
            input_file: A page image on which to perform OCR.
            output_hocr: The expected name of the output hOCR file.
            output_text: The expected name of a text file containing the
                recognized text.
            options: The command line options.
        """

    @staticmethod
    @abstractmethod
    def generate_pdf(
        input_file: Path, output_pdf: Path, output_text: Path, options: Namespace
    ) -> None:
        """Called to produce a text only PDF from a page image.

        A text only PDF should contain no visible material of any kind, as it
        will be grafted onto the input PDF page. It must be sized to the
        exact dimensions of the input image.

        This function executes in a worker thread or worker process. OCRmyPDF
        automatically parallelizes OCR over pages. The OCR engine should not
        introduce more parallelism.

        Args:
            input_file: A page image on which to perform OCR.
            output_pdf: The expected name of the output PDF.
            output_text: The expected name of a text file containing the
                recognized text.
            options: The command line options.
        """


@hookspec(firstresult=True)
def get_ocr_engine() -> OcrEngine:
    """Returns an OcrEngine to use for processing this file.

    The OcrEngine may be instantiated multiple times, by both the main process
    and child process.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


@hookspec(firstresult=True)
def generate_pdfa(
    pdf_pages: list[Path],
    pdfmark: Path,
    output_file: Path,
    context: PdfContext,
    pdf_version: str,
    pdfa_part: str,
    progressbar_class: type[ProgressBar] | None,
    stop_on_soft_error: bool,
) -> Path:
    """Generate a PDF/A.

    This API strongly assumes a PDF/A generator with Ghostscript's semantics.

    OCRmyPDF will modify the metadata and possibly linearize the PDF/A after it
    is generated.

    Arguments:
        pdf_pages: A list of one or more filenames, will be merged into output_file.
        pdfmark: A PostScript file intended for Ghostscript with details on
            how to perform the PDF/A conversion.
        output_file: The name of the desired output file.
        context: The current context.
        pdf_version: The minimum PDF version that the output file should be.
            At its own discretion, the PDF/A generator may raise the version,
            but should not lower it.
        pdfa_part: The desired PDF/A compliance level, such as ``'2B'``.
        progressbar_class: The class of a progress bar, which must implement
            the ProgressBar protocol. If None, no progress is reported.
        stop_on_soft_error: If there is an "soft error" such that PDF/A generation
            can proceed and produce a valid PDF/A, but output may be invalid or
            may not visually resemble the original, the implementer of this hook
            should raise a detailed exception. If ``False``, continue processing
            and report by logging it. If the hook cannot proceed, it should always
            raise an exception, regardless of this setting.

    Returns:
        Path: If successful, the hook should return ``output_file``.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.

    Note:
        Before version 15.0.0, the ``context`` was not provided and ``compression``
        was provided instead. Plugins should now read the context object to determine
        if compression is requested.
    """


@hookspec(firstresult=True)
def optimize_pdf(
    input_pdf: Path,
    output_pdf: Path,
    context: PdfContext,
    executor: Executor,
    linearize: bool,
) -> tuple[Path, Sequence[str]]:
    """Optimize a PDF after image, OCR and metadata processing.

    If the input_pdf is a PDF/A, the plugin should modify input_pdf in a way
    that preserves the PDF/A status, or report to the user when this is not possible.

    If the implementation fails to produce a smaller file than the input file, it
    should return input_pdf instead.

    A plugin that implements a new optimizer may need to suppress the built-in
    optimizer by implementing an ``initialize`` hook.

    Arguments:
        input_pdf: The input PDF, which has OCR added.
        output_pdf: The requested filename of the output PDF which should be created
            by this optimization hook.
        context: The current context.
        executor: An initialized executor which may be used during optimization,
            to distribute optimization tasks.
        linearize: If True, OCRmyPDF requires ``optimize_pdf`` to return a linearized,
            also known as fast web view PDF.

    Returns:
        Path: If optimization is successful, the hook should return ``output_file``.
            If optimization does not produce a smaller file, the hook should return
            ``input_file``.
        Sequence[str]: Any comments that the plugin wishes to report to the user,
            especially reasons it was not able to further optimize the file. For
            example, the plugin could report that a required third party was not
            installed, so a specific optimization was not attempted.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """


@hookspec(firstresult=True)
def is_optimization_enabled(context: PdfContext) -> bool:
    """For a given PdfContext, OCRmyPDF asks the plugin if optimization is enabled.

    An optimization plugin might be installed and active but could be disabled by
    user settings.

    If this returns False, OCRmyPDF will take certain actions to finalize the PDF.

    Returns:
        True if the plugin's optimization is enabled.

    Note:
        This is a :ref:`firstresult hook<firstresult>`.
    """



# SPDX-FileCopyrightText: 2025 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""Compare text in PDFs."""

from __future__ import annotations

from subprocess import run
from tempfile import NamedTemporaryFile
from typing import Annotated

import typer


def main(
    pdf1: Annotated[typer.FileBinaryRead, typer.Argument()],
    pdf2: Annotated[typer.FileBinaryRead, typer.Argument()],
    engine: Annotated[str, typer.Option()] = 'pdftotext',
):
    """Compare text in PDFs."""

    text1 = run(
        ['pdftotext', '-layout', '-', '-'], stdin=pdf1, capture_output=True, check=True
    )
    text2 = run(
        ['pdftotext', '-layout', '-', '-'], stdin=pdf2, capture_output=True, check=True
    )

    with NamedTemporaryFile() as f1, NamedTemporaryFile() as f2:
        f1.write(text1.stdout)
        f1.flush()
        f2.write(text2.stdout)
        f2.flush()
        diff = run(
            ['diff', '--color=always', '--side-by-side', f1.name, f2.name],
            capture_output=True,
        )
        run(['less', '-R'], input=diff.stdout, check=True)
        if text1.stdout.strip() != text2.stdout.strip():
            return 1

    return 0


if __name__ == '__main__':
    typer.run(main)



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0
"""Wrappers to manage subprocess calls."""

from __future__ import annotations

import logging
import os
import re
import sys
from collections.abc import Callable, Mapping, Sequence
from contextlib import suppress
from pathlib import Path
from subprocess import PIPE, STDOUT, CalledProcessError, CompletedProcess, Popen
from subprocess import run as subprocess_run

from packaging.version import Version

from ocrmypdf.exceptions import MissingDependencyError

# pylint: disable=logging-format-interpolation

log = logging.getLogger(__name__)

Args = Sequence[Path | str]
OsEnviron = os._Environ  # pylint: disable=protected-access


def run(
    args: Args,
    *,
    env: OsEnviron | None = None,
    logs_errors_to_stdout: bool = False,
    check: bool = False,
    **kwargs,
) -> CompletedProcess:
    """Wrapper around :py:func:`subprocess.run`.

    The main purpose of this wrapper is to log subprocess output in an orderly
    fashion that identifies the responsible subprocess. An additional
    task is that this function goes to greater lengths to find possible Windows
    locations of our dependencies when they are not on the system PATH.

    Arguments should be identical to ``subprocess.run``, except for following:

    Args:
        args: Positional arguments to pass to ``subprocess.run``.
        env: A set of environment variables. If None, the OS environment is used.
        logs_errors_to_stdout: If True, indicates that the process writes its error
            messages to stdout rather than stderr, so stdout should be logged
            if there is an error. If False, stderr is logged. Could be used with
            stderr=STDOUT, stdout=PIPE for example.
        check: If True, raise an exception if the process exits with a non-zero
            status code. If False, the return value will indicate success or failure.
        kwargs: Additional arguments to pass to ``subprocess.run``.
    """
    args, env, process_log, _text = _fix_process_args(args, env, kwargs)

    stderr = None
    stderr_name = 'stderr' if not logs_errors_to_stdout else 'stdout'
    try:
        proc = subprocess_run(args, env=env, check=check, **kwargs)
    except CalledProcessError as e:
        stderr = getattr(e, stderr_name, None)
        raise
    else:
        stderr = getattr(proc, stderr_name, None)
    finally:
        if process_log.isEnabledFor(logging.DEBUG) and stderr:
            with suppress(AttributeError, UnicodeDecodeError):
                stderr = stderr.decode('utf-8', 'replace')
            if logs_errors_to_stdout:
                process_log.debug("stdout/stderr = %s", stderr)
            else:
                process_log.debug("stderr = %s", stderr)
    return proc


def run_polling_stderr(
    args: Args,
    *,
    callback: Callable[[str], None],
    check: bool = False,
    env: OsEnviron | None = None,
    **kwargs,
) -> CompletedProcess:
    """Run a process like ``ocrmypdf.subprocess.run``, and poll stderr.

    Every line of produced by stderr will be forwarded to the callback function.
    The intended use is monitoring progress of subprocesses that output their
    own progress indicators. In addition, each line will be logged if debug
    logging is enabled.

    Requires stderr to be opened in text mode for ease of handling errors. In
    addition the expected encoding= and errors= arguments should be set. Note
    that if stdout is already set up, it need not be binary.
    """
    args, env, process_log, text = _fix_process_args(args, env, kwargs)
    assert text, "Must use text=True"

    with Popen(args, env=env, **kwargs) as proc:
        lines = []
        while proc.poll() is None:
            if proc.stderr is None:
                continue
            for msg in iter(proc.stderr.readline, ''):
                if process_log.isEnabledFor(logging.DEBUG):
                    process_log.debug(msg.strip())
                callback(msg)
                lines.append(msg)
        stderr = ''.join(lines)

        if check and proc.returncode != 0:
            raise CalledProcessError(proc.returncode, args, output=None, stderr=stderr)
        return CompletedProcess(args, proc.returncode, None, stderr=stderr)


def _fix_process_args(
    args: Args, env: OsEnviron | None, kwargs
) -> tuple[Args, OsEnviron, logging.Logger, bool]:
    if not env:
        env = os.environ

    # Search in spoof path if necessary
    program = str(args[0])

    if sys.platform == 'win32':
        # pylint: disable=import-outside-toplevel
        from ocrmypdf.subprocess._windows import fix_windows_args

        args = fix_windows_args(program, args, env)

    log.debug("Running: %s", args)
    process_log = log.getChild(os.path.basename(program))
    text = bool(kwargs.get('text', False))

    return args, env, process_log, text


def get_version(
    program: str,
    *,
    version_arg: str = '--version',
    regex=r'(\d+(\.\d+)*)',
    env: OsEnviron | None = None,
) -> str:
    """Get the version of the specified program.

    Arguments:
        program: The program to version check.
        version_arg: The argument needed to ask for its version, e.g. ``--version``.
        regex: A regular expression to parse the program's output and obtain the
            version.
        env: Custom ``os.environ`` in which to run program.
    """
    args_prog = [program, version_arg]
    try:
        proc = run(
            args_prog,
            close_fds=True,
            text=True,
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
            env=env,
        )
        output: str = proc.stdout
    except FileNotFoundError as e:
        raise MissingDependencyError(
            f"Could not find program '{program}' on the PATH"
        ) from e
    except CalledProcessError as e:
        if e.returncode != 0:
            log.exception(e)
            raise MissingDependencyError(
                f"Ran program '{program}' but it exited with an error:\n{e.output}"
            ) from e
        raise MissingDependencyError(
            f"Could not find program '{program}' on the PATH"
        ) from e

    match = re.match(regex, output.strip())
    if not match:
        raise MissingDependencyError(
            f"The program '{program}' did not report its version. "
            f"Message was:\n{output}"
        )
    version = match.group(1)

    return version


MISSING_PROGRAM = '''
The program '{program}' could not be executed or was not found on your
system PATH.
'''

MISSING_OPTIONAL_PROGRAM = '''
The program '{program}' could not be executed or was not found on your
system PATH.  This program is required when you use the
{required_for} arguments.  You could try omitting these arguments, or install
the package.
'''

MISSING_RECOMMEND_PROGRAM = '''
The program '{program}' could not be executed or was not found on your
system PATH.  This program is recommended when using the {required_for} arguments,
but not required, so we will proceed.  For best results, install the program.
'''

OLD_VERSION = '''
OCRmyPDF requires '{program}' {need_version} or higher.  Your system appears
to have {found_version}.  Please update this program.
'''

OLD_VERSION_REQUIRED_FOR = '''
OCRmyPDF requires '{program}' {need_version} or higher when run with the
{required_for} arguments.  {program} {found_version} is installed.

If you omit these arguments, OCRmyPDF may be able to
proceed.  For best results, update the program.
'''

OSX_INSTALL_ADVICE = '''
If you have homebrew installed, try these command to install the missing
package:
    brew install {package}
'''

LINUX_INSTALL_ADVICE = '''
On systems with the aptitude package manager (Debian, Ubuntu), try these
commands:
    sudo apt update
    sudo apt install {package}

On RPM-based systems (Red Hat, Fedora), try this command:
    sudo dnf install {package}
'''

WINDOWS_INSTALL_ADVICE = '''
If not already installed, install the Chocolatey package manager. Then use
a command prompt to install the missing package:
    choco install {package}
'''


def _get_platform() -> str:
    if sys.platform.startswith('freebsd'):
        return 'freebsd'
    elif sys.platform.startswith('linux'):
        return 'linux'
    elif sys.platform.startswith('win'):
        return 'windows'
    return sys.platform


def _error_trailer(program: str, package: str | Mapping[str, str], **kwargs) -> None:
    del kwargs
    if isinstance(package, Mapping):
        package = package.get(_get_platform(), program)

    if _get_platform() == 'darwin':
        log.info(OSX_INSTALL_ADVICE.format(**locals()))
    elif _get_platform() == 'linux':
        log.info(LINUX_INSTALL_ADVICE.format(**locals()))
    elif _get_platform() == 'windows':
        log.info(WINDOWS_INSTALL_ADVICE.format(**locals()))


def _error_missing_program(
    program: str, package: str, required_for: str | None, recommended: bool
) -> None:
    # pylint: disable=unused-argument
    if recommended:
        log.warning(MISSING_RECOMMEND_PROGRAM.format(**locals()))
    elif required_for:
        log.error(MISSING_OPTIONAL_PROGRAM.format(**locals()))
    else:
        log.error(MISSING_PROGRAM.format(**locals()))
    _error_trailer(**locals())


def _error_old_version(
    program: str,
    package: str,
    need_version: str,
    found_version: str,
    required_for: str | None,
) -> None:
    # pylint: disable=unused-argument
    if required_for:
        log.error(OLD_VERSION_REQUIRED_FOR.format(**locals()))
    else:
        log.error(OLD_VERSION.format(**locals()))
    _error_trailer(**locals())


def check_external_program(
    *,
    program: str,
    package: str,
    version_checker: Callable[[], Version],
    need_version: str | Version,
    required_for: str | None = None,
    recommended: bool = False,
    version_parser: type[Version] = Version,
) -> None:
    """Check for required version of external program and raise exception if not.

    Args:
        program: The name of the program to test.
        package: The name of a software package that typically supplies this program.
            Usually the same as program.
        version_checker: A callable without arguments that retrieves the installed
            version of program.
        need_version: The minimum required version.
        required_for: The name of an argument of feature that requires this program.
        recommended: If this external program is recommended, instead of raising
            an exception, log a warning and allow execution to continue.
        version_parser: A class that should be used to parse and compare version
            numbers. Used when version numbers do not follow standard conventions.
    """
    if not isinstance(need_version, Version):
        need_version = version_parser(need_version)
    try:
        found_version = version_checker()
    except (CalledProcessError, FileNotFoundError) as e:
        _error_missing_program(program, package, required_for, recommended)
        if not recommended:
            raise MissingDependencyError(program) from e
        return
    except MissingDependencyError:
        _error_missing_program(program, package, required_for, recommended)
        if not recommended:
            raise
        return

    if found_version and found_version < need_version:
        _error_old_version(
            program, package, str(need_version), str(found_version), required_for
        )
        if not recommended:
            raise MissingDependencyError(program)

    log.debug('Found %s %s', program, found_version)



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MIT
from __future__ import annotations

from contextlib import contextmanager
from subprocess import CalledProcessError
from unittest.mock import patch

from ocrmypdf import hookimpl
from ocrmypdf.builtin_plugins.tesseract_ocr import TesseractOcrEngine


def raise_size_exception(*args, **kwargs):
    raise CalledProcessError(
        1,
        'tesseract',
        output=b"Image too large: (33830, 14959)\nError during processing.",
        stderr=b"",
    )


@contextmanager
def patch_tesseract_run():
    with patch('ocrmypdf._exec.tesseract.run') as mock:
        mock.side_effect = raise_size_exception
        yield
        mock.assert_called()


class BigImageErrorOcrEngine(TesseractOcrEngine):
    @staticmethod
    def get_orientation(input_file, options):
        with patch_tesseract_run():
            return TesseractOcrEngine.get_orientation(input_file, options)

    @staticmethod
    def generate_hocr(input_file, output_hocr, output_text, options):
        with patch_tesseract_run():
            TesseractOcrEngine.generate_hocr(
                input_file, output_hocr, output_text, options
            )

    @staticmethod
    def generate_pdf(input_file, output_pdf, output_text, options):
        with patch_tesseract_run():
            TesseractOcrEngine.generate_pdf(
                input_file, output_pdf, output_text, options
            )


@hookimpl
def get_ocr_engine():
    return BigImageErrorOcrEngine()



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MIT

from __future__ import annotations

from subprocess import CalledProcessError
from unittest.mock import patch

from ocrmypdf import hookimpl
from ocrmypdf.builtin_plugins import ghostscript
from ocrmypdf.subprocess import run_polling_stderr


def fail_if_stoponerror(args, **kwargs):
    if '-dPDFSTOPONERROR' in args:
        raise CalledProcessError(1, 'gs', output=b"", stderr=b"PDF STOP ON ERROR")
    return run_polling_stderr(args, **kwargs)


@hookimpl
def generate_pdfa(
    pdf_pages,
    pdfmark,
    output_file,
    context,
    pdf_version,
    pdfa_part,
    stop_on_soft_error,
):
    with patch('ocrmypdf._exec.ghostscript.run_polling_stderr') as mock:
        mock.side_effect = fail_if_stoponerror
        ghostscript.generate_pdfa(
            pdf_pages=pdf_pages,
            pdfmark=pdfmark,
            output_file=output_file,
            context=context,
            pdf_version=pdf_version,
            pdfa_part=pdfa_part,
            progressbar_class=None,
            stop_on_soft_error=stop_on_soft_error,
        )
        mock.assert_called()
        return output_file



# SPDX-FileCopyrightText: 2025 James R. Barlow
# SPDX-License-Identifier: MIT

"""Run OCRmyPDF on the same PDF with different options."""

from __future__ import annotations

import os
import shlex
from io import BytesIO
from pathlib import Path
from subprocess import check_output, run
from tempfile import TemporaryDirectory

import pikepdf
import pymupdf
import streamlit as st
from lxml import etree
from streamlit_pdf_viewer import pdf_viewer


def do_column(label, suffix, d):
    cli = st.text_area(
        f"Command line arguments for {label}",
        key=f"args{suffix}",
        value="ocrmypdf {in_} {out}",
    )
    env_text = st.text_area(f"Environment variables for {label}", key=f"env{suffix}")
    env = os.environ.copy()
    for line in env_text.splitlines():
        if line:
            try:
                k, v = line.split("=", 1)
            except ValueError:
                st.error(f"Invalid environment variable: {line}")
                break
            env[k] = v
    args = shlex.split(
        cli.format(
            in_=os.path.join(d, "input.pdf"),
            out=os.path.join(d, f"output{suffix}.pdf"),
        )
    )
    with st.expander("Environment variables", expanded=bool(env_text.strip())):
        st.code('\n'.join(f"{k}={v}" for k, v in env.items()))
    st.code(shlex.join(args))
    return env, args


def main():
    st.set_page_config(layout="wide")

    st.title("OCRmyPDF Compare")
    st.write("Run OCRmyPDF on the same PDF with different options.")
    st.warning("This is a testing tool and is not intended for production use.")

    uploaded_pdf = st.file_uploader("Upload a PDF", type=["pdf"])
    if uploaded_pdf is None:
        return

    pdf_bytes = uploaded_pdf.read()

    with pikepdf.open(BytesIO(pdf_bytes)) as p, TemporaryDirectory() as d:
        with st.expander("PDF Metadata"):
            with p.open_metadata() as meta:
                xml_txt = str(meta)
                parser = etree.XMLParser(remove_blank_text=True)
                tree = etree.fromstring(xml_txt, parser=parser)
                st.code(
                    etree.tostring(tree, pretty_print=True).decode("utf-8"),
                    language="xml",
                )
            st.write(p.docinfo)
            st.write("Number of pages:", len(p.pages))

        col1, col2 = st.columns(2)
        with col1:
            env1, args1 = do_column("A", "1", d)
        with col2:
            env2, args2 = do_column("B", "2", d)

        if not st.button("Execute and Compare"):
            return
        with st.spinner("Executing..."):
            Path(d, "input.pdf").write_bytes(pdf_bytes)
            run(args1, env=env1)
            run(args2, env=env2)

            col1, col2 = st.columns(2)
            with col1:
                st.text(
                    "Ghostscript version A: "
                    + check_output(
                        ["gs", "--version"],
                        env=env1,
                        text=True,
                    )
                )
            with col2:
                st.text(
                    "Ghostscript version B: "
                    + check_output(
                        ["gs", "--version"],
                        env=env2,
                        text=True,
                    )
                )

            doc1 = pymupdf.open(os.path.join(d, "output1.pdf"))
            doc2 = pymupdf.open(os.path.join(d, "output2.pdf"))
            for i, page1_2 in enumerate(zip(doc1, doc2)):
                st.write(f"Page {i+1}")
                page1, page2 = page1_2
                col1, col2 = st.columns(2)
                with col1, st.container(border=True):
                    st.write(page1.get_text())
                with col2, st.container(border=True):
                    st.write(page2.get_text())

            col1, col2 = st.columns(2)
            with col1, st.expander("PDF Viewer"):
                pdf_viewer(Path(d, "output1.pdf"))
            with col2, st.expander("PDF Viewer"):
                pdf_viewer(Path(d, "output2.pdf"))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""Language codes and names from ISO 639.

Derived from
https://www.loc.gov/standards/iso639-2/ascii_8bits.html
"""

from typing import NamedTuple


class ISOCodeData(NamedTuple):
    """Data for a single ISO 639 code."""

    alt: str
    alpha_2: str
    english: str
    french: str


ISO_639_3 = {
    'aar': ISOCodeData('', 'aa', 'Afar', 'afar'),
    'abk': ISOCodeData('', 'ab', 'Abkhazian', 'abkhaze'),
    'ace': ISOCodeData('', '', 'Achinese', 'aceh'),
    'ach': ISOCodeData('', '', 'Acoli', 'acoli'),
    'ada': ISOCodeData('', '', 'Adangme', 'adangme'),
    'ady': ISOCodeData('', '', 'Adyghe; Adygei', 'adyghé'),
    'afa': ISOCodeData(
        '',
        '',
        'Afro-Asiatic languages',
        'afro-asiatiques, langues',
    ),
    'afh': ISOCodeData('', '', 'Afrihili', 'afrihili'),
    'afr': ISOCodeData('', 'af', 'Afrikaans', 'afrikaans'),
    'ain': ISOCodeData('', '', 'Ainu', 'aïnou'),
    'aka': ISOCodeData('', 'ak', 'Akan', 'akan'),
    'akk': ISOCodeData('', '', 'Akkadian', 'akkadien'),
    'alb': ISOCodeData('sqi', 'sq', 'Albanian', 'albanais'),
    'ale': ISOCodeData('', '', 'Aleut', 'aléoute'),
    'alg': ISOCodeData(
        '',
        '',
        'Algonquian languages',
        'algonquines, langues',
    ),
    'alt': ISOCodeData('', '', 'Southern Altai', 'altai du Sud'),
    'amh': ISOCodeData('', 'am', 'Amharic', 'amharique'),
    'ang': ISOCodeData(
        '',
        '',
        'English, Old (ca.450-1100)',
        'anglo-saxon (ca.450-1100)',
    ),
    'anp': ISOCodeData('', '', 'Angika', 'angika'),
    'apa': ISOCodeData('', '', 'Apache languages', 'apaches, langues'),
    'ara': ISOCodeData('', 'ar', 'Arabic', 'arabe'),
    'arc': ISOCodeData(
        '',
        '',
        'Official Aramaic (700-300 BCE); Imperial Aramaic (700-300 BCE)',
        "araméen d'empire (700-300 BCE)",
    ),
    'arg': ISOCodeData('', 'an', 'Aragonese', 'aragonais'),
    'arm': ISOCodeData('hye', 'hy', 'Armenian', 'arménien'),
    'arn': ISOCodeData(
        '',
        '',
        'Mapudungun; Mapuche',
        'mapudungun; mapuche; mapuce',
    ),
    'arp': ISOCodeData('', '', 'Arapaho', 'arapaho'),
    'art': ISOCodeData(
        '',
        '',
        'Artificial languages',
        'artificielles, langues',
    ),
    'arw': ISOCodeData('', '', 'Arawak', 'arawak'),
    'asm': ISOCodeData('', 'as', 'Assamese', 'assamais'),
    'ast': ISOCodeData(
        '',
        '',
        'Asturian; Bable; Leonese; Asturleonese',
        'asturien; bable; léonais; asturoléonais',
    ),
    'ath': ISOCodeData(
        '',
        '',
        'Athapascan languages',
        'athapascanes, langues',
    ),
    'aus': ISOCodeData(
        '',
        '',
        'Australian languages',
        'australiennes, langues',
    ),
    'ava': ISOCodeData('', 'av', 'Avaric', 'avar'),
    'ave': ISOCodeData('', 'ae', 'Avestan', 'avestique'),
    'awa': ISOCodeData('', '', 'Awadhi', 'awadhi'),
    'aym': ISOCodeData('', 'ay', 'Aymara', 'aymara'),
    'aze': ISOCodeData('', 'az', 'Azerbaijani', 'azéri'),
    'bad': ISOCodeData('', '', 'Banda languages', 'banda, langues'),
    'bai': ISOCodeData('', '', 'Bamileke languages', 'bamiléké, langues'),
    'bak': ISOCodeData('', 'ba', 'Bashkir', 'bachkir'),
    'bal': ISOCodeData('', '', 'Baluchi', 'baloutchi'),
    'bam': ISOCodeData('', 'bm', 'Bambara', 'bambara'),
    'ban': ISOCodeData('', '', 'Balinese', 'balinais'),
    'baq': ISOCodeData('eus', 'eu', 'Basque', 'basque'),
    'bas': ISOCodeData('', '', 'Basa', 'basa'),
    'bat': ISOCodeData('', '', 'Baltic languages', 'baltes, langues'),
    'bej': ISOCodeData('', '', 'Beja; Bedawiyet', 'bedja'),
    'bel': ISOCodeData('', 'be', 'Belarusian', 'biélorusse'),
    'bem': ISOCodeData('', '', 'Bemba', 'bemba'),
    'ben': ISOCodeData('', 'bn', 'Bengali', 'bengali'),
    'ber': ISOCodeData('', '', 'Berber languages', 'berbères, langues'),
    'bho': ISOCodeData('', '', 'Bhojpuri', 'bhojpuri'),
    'bih': ISOCodeData('', 'bh', 'Bihari languages', 'langues biharis'),
    'bik': ISOCodeData('', '', 'Bikol', 'bikol'),
    'bin': ISOCodeData('', '', 'Bini; Edo', 'bini; edo'),
    'bis': ISOCodeData('', 'bi', 'Bislama', 'bichlamar'),
    'bla': ISOCodeData('', '', 'Siksika', 'blackfoot'),
    'bnt': ISOCodeData('', '', 'Bantu languages', 'bantou, langues'),
    'bos': ISOCodeData('', 'bs', 'Bosnian', 'bosniaque'),
    'bra': ISOCodeData('', '', 'Braj', 'braj'),
    'bre': ISOCodeData('', 'br', 'Breton', 'breton'),
    'btk': ISOCodeData('', '', 'Batak languages', 'batak, langues'),
    'bua': ISOCodeData('', '', 'Buriat', 'bouriate'),
    'bug': ISOCodeData('', '', 'Buginese', 'bugi'),
    'bul': ISOCodeData('', 'bg', 'Bulgarian', 'bulgare'),
    'bur': ISOCodeData('mya', 'my', 'Burmese', 'birman'),
    'byn': ISOCodeData('', '', 'Blin; Bilin', 'blin; bilen'),
    'cad': ISOCodeData('', '', 'Caddo', 'caddo'),
    'cai': ISOCodeData(
        '',
        '',
        'Central American Indian languages',
        "amérindiennes de L'Amérique centrale, langues",
    ),
    'car': ISOCodeData('', '', 'Galibi Carib', 'karib; galibi; carib'),
    'cat': ISOCodeData('', 'ca', 'Catalan; Valencian', 'catalan; valencien'),
    'cau': ISOCodeData(
        '',
        '',
        'Caucasian languages',
        'caucasiennes, langues',
    ),
    'ceb': ISOCodeData('', '', 'Cebuano', 'cebuano'),
    'cel': ISOCodeData(
        '',
        '',
        'Celtic languages',
        'celtiques, langues; celtes, langues',
    ),
    'cha': ISOCodeData('', 'ch', 'Chamorro', 'chamorro'),
    'chb': ISOCodeData('', '', 'Chibcha', 'chibcha'),
    'che': ISOCodeData('', 'ce', 'Chechen', 'tchétchène'),
    'chg': ISOCodeData('', '', 'Chagatai', 'djaghataï'),
    'chi': ISOCodeData('zho', 'zh', 'Chinese', 'chinois'),
    'chk': ISOCodeData('', '', 'Chuukese', 'chuuk'),
    'chm': ISOCodeData('', '', 'Mari', 'mari'),
    'chn': ISOCodeData('', '', 'Chinook jargon', 'chinook, jargon'),
    'cho': ISOCodeData('', '', 'Choctaw', 'choctaw'),
    'chp': ISOCodeData('', '', 'Chipewyan; Dene Suline', 'chipewyan'),
    'chr': ISOCodeData('', '', 'Cherokee', 'cherokee'),
    'chu': ISOCodeData(
        '',
        'cu',
        (
            'Church Slavic; Old Slavonic; Church Slavonic;'
            ' Old Bulgarian; Old Church Slavonic'
        ),
        "slavon d'église; vieux slave; slavon liturgique; vieux bulgare",
    ),
    'chv': ISOCodeData('', 'cv', 'Chuvash', 'tchouvache'),
    'chy': ISOCodeData('', '', 'Cheyenne', 'cheyenne'),
    'cmc': ISOCodeData('', '', 'Chamic languages', 'chames, langues'),
    'cnr': ISOCodeData('', '', 'Montenegrin', 'monténégrin'),
    'cop': ISOCodeData('', '', 'Coptic', 'copte'),
    'cor': ISOCodeData('', 'kw', 'Cornish', 'cornique'),
    'cos': ISOCodeData('', 'co', 'Corsican', 'corse'),
    'cpe': ISOCodeData(
        '',
        '',
        'Creoles and pidgins, English based',
        "créoles et pidgins basés sur l'anglais",
    ),
    'cpf': ISOCodeData(
        '',
        '',
        'Creoles and pidgins, French-based',
        'créoles et pidgins basés sur le français',
    ),
    'cpp': ISOCodeData(
        '',
        '',
        'Creoles and pidgins, Portuguese-based',
        'créoles et pidgins basés sur le portugais',
    ),
    'cre': ISOCodeData('', 'cr', 'Cree', 'cree'),
    'crh': ISOCodeData(
        '',
        '',
        'Crimean Tatar; Crimean Turkish',
        'tatar de Crimé',
    ),
    'crp': ISOCodeData('', '', 'Creoles and pidgins', 'créoles et pidgins'),
    'csb': ISOCodeData('', '', 'Kashubian', 'kachoube'),
    'cus': ISOCodeData('', '', 'Cushitic languages', 'couchitiques, langues'),
    'cze': ISOCodeData('ces', 'cs', 'Czech', 'tchèque'),
    'dak': ISOCodeData('', '', 'Dakota', 'dakota'),
    'dan': ISOCodeData('', 'da', 'Danish', 'danois'),
    'dar': ISOCodeData('', '', 'Dargwa', 'dargwa'),
    'day': ISOCodeData('', '', 'Land Dayak languages', 'dayak, langues'),
    'del': ISOCodeData('', '', 'Delaware', 'delaware'),
    'den': ISOCodeData('', '', 'Slave (Athapascan)', 'esclave (athapascan)'),
    'dgr': ISOCodeData('', '', 'Dogrib', 'dogrib'),
    'din': ISOCodeData('', '', 'Dinka', 'dinka'),
    'div': ISOCodeData('', 'dv', 'Divehi; Dhivehi; Maldivian', 'maldivien'),
    'doi': ISOCodeData('', '', 'Dogri', 'dogri'),
    'dra': ISOCodeData(
        '',
        '',
        'Dravidian languages',
        'dravidiennes, langues',
    ),
    'dsb': ISOCodeData('', '', 'Lower Sorbian', 'bas-sorabe'),
    'dua': ISOCodeData('', '', 'Duala', 'douala'),
    'dum': ISOCodeData(
        '',
        '',
        'Dutch, Middle (ca.1050-1350)',
        'néerlandais moyen (ca. 1050-1350)',
    ),
    'dut': ISOCodeData('nld', 'nl', 'Dutch; Flemish', 'néerlandais; flamand'),
    'dyu': ISOCodeData('', '', 'Dyula', 'dioula'),
    'dzo': ISOCodeData('', 'dz', 'Dzongkha', 'dzongkha'),
    'efi': ISOCodeData('', '', 'Efik', 'efik'),
    'egy': ISOCodeData('', '', 'Egyptian (Ancient)', 'égyptien'),
    'eka': ISOCodeData('', '', 'Ekajuk', 'ekajuk'),
    'elx': ISOCodeData('', '', 'Elamite', 'élamite'),
    'eng': ISOCodeData('', 'en', 'English', 'anglais'),
    'enm': ISOCodeData(
        '',
        '',
        'English, Middle (1100-1500)',
        'anglais moyen (1100-1500)',
    ),
    'epo': ISOCodeData('', 'eo', 'Esperanto', 'espéranto'),
    'est': ISOCodeData('', 'et', 'Estonian', 'estonien'),
    'ewe': ISOCodeData('', 'ee', 'Ewe', 'éwé'),
    'ewo': ISOCodeData('', '', 'Ewondo', 'éwondo'),
    'fan': ISOCodeData('', '', 'Fang', 'fang'),
    'fao': ISOCodeData('', 'fo', 'Faroese', 'féroïen'),
    'fat': ISOCodeData('', '', 'Fanti', 'fanti'),
    'fij': ISOCodeData('', 'fj', 'Fijian', 'fidjien'),
    'fil': ISOCodeData('', '', 'Filipino; Pilipino', 'filipino; pilipino'),
    'fin': ISOCodeData('', 'fi', 'Finnish', 'finnois'),
    'fiu': ISOCodeData(
        '',
        '',
        'Finno-Ugrian languages',
        'finno-ougriennes, langues',
    ),
    'fon': ISOCodeData('', '', 'Fon', 'fon'),
    'fre': ISOCodeData('fra', 'fr', 'French', 'français'),
    'frm': ISOCodeData(
        '',
        '',
        'French, Middle (ca.1400-1600)',
        'français moyen (1400-1600)',
    ),
    'fro': ISOCodeData(
        '',
        '',
        'French, Old (842-ca.1400)',
        'français ancien (842-ca.1400)',
    ),
    'frr': ISOCodeData('', '', 'Northern Frisian', 'frison septentrional'),
    'frs': ISOCodeData('', '', 'Eastern Frisian', 'frison oriental'),
    'fry': ISOCodeData('', 'fy', 'Western Frisian', 'frison occidental'),
    'ful': ISOCodeData('', 'ff', 'Fulah', 'peul'),
    'fur': ISOCodeData('', '', 'Friulian', 'frioulan'),
    'gaa': ISOCodeData('', '', 'Ga', 'ga'),
    'gay': ISOCodeData('', '', 'Gayo', 'gayo'),
    'gba': ISOCodeData('', '', 'Gbaya', 'gbaya'),
    'gem': ISOCodeData('', '', 'Germanic languages', 'germaniques, langues'),
    'geo': ISOCodeData('kat', 'ka', 'Georgian', 'géorgien'),
    'ger': ISOCodeData('deu', 'de', 'German', 'allemand'),
    'gez': ISOCodeData('', '', 'Geez', 'guèze'),
    'gil': ISOCodeData('', '', 'Gilbertese', 'kiribati'),
    'gla': ISOCodeData(
        '',
        'gd',
        'Gaelic; Scottish Gaelic',
        'gaélique; gaélique écossais',
    ),
    'gle': ISOCodeData('', 'ga', 'Irish', 'irlandais'),
    'glg': ISOCodeData('', 'gl', 'Galician', 'galicien'),
    'glv': ISOCodeData('', 'gv', 'Manx', 'manx; mannois'),
    'gmh': ISOCodeData(
        '',
        '',
        'German, Middle High (ca.1050-1500)',
        'allemand, moyen haut (ca. 1050-1500)',
    ),
    'goh': ISOCodeData(
        '',
        '',
        'German, Old High (ca.750-1050)',
        'allemand, vieux haut (ca. 750-1050)',
    ),
    'gon': ISOCodeData('', '', 'Gondi', 'gond'),
    'gor': ISOCodeData('', '', 'Gorontalo', 'gorontalo'),
    'got': ISOCodeData('', '', 'Gothic', 'gothique'),
    'grb': ISOCodeData('', '', 'Grebo', 'grebo'),
    'grc': ISOCodeData(
        '',
        '',
        'Greek, Ancient (to 1453)',
        "grec ancien (jusqu'à 1453)",
    ),
    'gre': ISOCodeData(
        'ell',
        'el',
        'Greek, Modern (1453-)',
        'grec moderne (après 1453)',
    ),
    'grn': ISOCodeData('', 'gn', 'Guarani', 'guarani'),
    'gsw': ISOCodeData(
        '',
        '',
        'Swiss German; Alemannic; Alsatian',
        'suisse alémanique; alémanique; alsacien',
    ),
    'guj': ISOCodeData('', 'gu', 'Gujarati', 'goudjrati'),
    'gwi': ISOCodeData('', '', "Gwich'in", "gwich'in"),
    'hai': ISOCodeData('', '', 'Haida', 'haida'),
    'hat': ISOCodeData(
        '',
        'ht',
        'Haitian; Haitian Creole',
        'haïtien; créole haïtien',
    ),
    'hau': ISOCodeData('', 'ha', 'Hausa', 'haoussa'),
    'haw': ISOCodeData('', '', 'Hawaiian', 'hawaïen'),
    'heb': ISOCodeData('', 'he', 'Hebrew', 'hébreu'),
    'her': ISOCodeData('', 'hz', 'Herero', 'herero'),
    'hil': ISOCodeData('', '', 'Hiligaynon', 'hiligaynon'),
    'him': ISOCodeData(
        '',
        '',
        'Himachali languages; Western Pahari languages',
        'langues himachalis; langues paharis occidentales',
    ),
    'hin': ISOCodeData('', 'hi', 'Hindi', 'hindi'),
    'hit': ISOCodeData('', '', 'Hittite', 'hittite'),
    'hmn': ISOCodeData('', '', 'Hmong; Mong', 'hmong'),
    'hmo': ISOCodeData('', 'ho', 'Hiri Motu', 'hiri motu'),
    'hrv': ISOCodeData('', 'hr', 'Croatian', 'croate'),
    'hsb': ISOCodeData('', '', 'Upper Sorbian', 'haut-sorabe'),
    'hun': ISOCodeData('', 'hu', 'Hungarian', 'hongrois'),
    'hup': ISOCodeData('', '', 'Hupa', 'hupa'),
    'iba': ISOCodeData('', '', 'Iban', 'iban'),
    'ibo': ISOCodeData('', 'ig', 'Igbo', 'igbo'),
    'ice': ISOCodeData('isl', 'is', 'Icelandic', 'islandais'),
    'ido': ISOCodeData('', 'io', 'Ido', 'ido'),
    'iii': ISOCodeData('', 'ii', 'Sichuan Yi; Nuosu', 'yi de Sichuan'),
    'ijo': ISOCodeData('', '', 'Ijo languages', 'ijo, langues'),
    'iku': ISOCodeData('', 'iu', 'Inuktitut', 'inuktitut'),
    'ile': ISOCodeData('', 'ie', 'Interlingue; Occidental', 'interlingue'),
    'ilo': ISOCodeData('', '', 'Iloko', 'ilocano'),
    'ina': ISOCodeData(
        '',
        'ia',
        'Interlingua (International Auxiliary Language Association)',
        'interlingua (langue auxiliaire internationale)',
    ),
    'inc': ISOCodeData('', '', 'Indic languages', 'indo-aryennes, langues'),
    'ind': ISOCodeData('', 'id', 'Indonesian', 'indonésien'),
    'ine': ISOCodeData(
        '',
        '',
        'Indo-European languages',
        'indo-européennes, langues',
    ),
    'inh': ISOCodeData('', '', 'Ingush', 'ingouche'),
    'ipk': ISOCodeData('', 'ik', 'Inupiaq', 'inupiaq'),
    'ira': ISOCodeData('', '', 'Iranian languages', 'iraniennes, langues'),
    'iro': ISOCodeData('', '', 'Iroquoian languages', 'iroquoises, langues'),
    'ita': ISOCodeData('', 'it', 'Italian', 'italien'),
    'jav': ISOCodeData('', 'jv', 'Javanese', 'javanais'),
    'jbo': ISOCodeData('', '', 'Lojban', 'lojban'),
    'jpn': ISOCodeData('', 'ja', 'Japanese', 'japonais'),
    'jpr': ISOCodeData('', '', 'Judeo-Persian', 'judéo-persan'),
    'jrb': ISOCodeData('', '', 'Judeo-Arabic', 'judéo-arabe'),
    'kaa': ISOCodeData('', '', 'Kara-Kalpak', 'karakalpak'),
    'kab': ISOCodeData('', '', 'Kabyle', 'kabyle'),
    'kac': ISOCodeData('', '', 'Kachin; Jingpho', 'kachin; jingpho'),
    'kal': ISOCodeData('', 'kl', 'Kalaallisut; Greenlandic', 'groenlandais'),
    'kam': ISOCodeData('', '', 'Kamba', 'kamba'),
    'kan': ISOCodeData('', 'kn', 'Kannada', 'kannada'),
    'kar': ISOCodeData('', '', 'Karen languages', 'karen, langues'),
    'kas': ISOCodeData('', 'ks', 'Kashmiri', 'kashmiri'),
    'kau': ISOCodeData('', 'kr', 'Kanuri', 'kanouri'),
    'kaw': ISOCodeData('', '', 'Kawi', 'kawi'),
    'kaz': ISOCodeData('', 'kk', 'Kazakh', 'kazakh'),
    'kbd': ISOCodeData('', '', 'Kabardian', 'kabardien'),
    'kha': ISOCodeData('', '', 'Khasi', 'khasi'),
    'khi': ISOCodeData('', '', 'Khoisan languages', 'khoïsan, langues'),
    'khm': ISOCodeData('', 'km', 'Central Khmer', 'khmer central'),
    'kho': ISOCodeData('', '', 'Khotanese; Sakan', 'khotanais; sakan'),
    'kik': ISOCodeData('', 'ki', 'Kikuyu; Gikuyu', 'kikuyu'),
    'kin': ISOCodeData('', 'rw', 'Kinyarwanda', 'rwanda'),
    'kir': ISOCodeData('', 'ky', 'Kirghiz; Kyrgyz', 'kirghiz'),
    'kmb': ISOCodeData('', '', 'Kimbundu', 'kimbundu'),
    'kok': ISOCodeData('', '', 'Konkani', 'konkani'),
    'kom': ISOCodeData('', 'kv', 'Komi', 'kom'),
    'kon': ISOCodeData('', 'kg', 'Kongo', 'kongo'),
    'kor': ISOCodeData('', 'ko', 'Korean', 'coréen'),
    'kos': ISOCodeData('', '', 'Kosraean', 'kosrae'),
    'kpe': ISOCodeData('', '', 'Kpelle', 'kpellé'),
    'krc': ISOCodeData('', '', 'Karachay-Balkar', 'karatchai balkar'),
    'krl': ISOCodeData('', '', 'Karelian', 'carélien'),
    'kro': ISOCodeData('', '', 'Kru languages', 'krou, langues'),
    'kru': ISOCodeData('', '', 'Kurukh', 'kurukh'),
    'kua': ISOCodeData('', 'kj', 'Kuanyama; Kwanyama', 'kuanyama; kwanyama'),
    'kum': ISOCodeData('', '', 'Kumyk', 'koumyk'),
    'kur': ISOCodeData('', 'ku', 'Kurdish', 'kurde'),
    'kut': ISOCodeData('', '', 'Kutenai', 'kutenai'),
    'lad': ISOCodeData('', '', 'Ladino', 'judéo-espagnol'),
    'lah': ISOCodeData('', '', 'Lahnda', 'lahnda'),
    'lam': ISOCodeData('', '', 'Lamba', 'lamba'),
    'lao': ISOCodeData('', 'lo', 'Lao', 'lao'),
    'lat': ISOCodeData('', 'la', 'Latin', 'latin'),
    'lav': ISOCodeData('', 'lv', 'Latvian', 'letton'),
    'lez': ISOCodeData('', '', 'Lezghian', 'lezghien'),
    'lim': ISOCodeData(
        '',
        'li',
        'Limburgan; Limburger; Limburgish',
        'limbourgeois',
    ),
    'lin': ISOCodeData('', 'ln', 'Lingala', 'lingala'),
    'lit': ISOCodeData('', 'lt', 'Lithuanian', 'lituanien'),
    'lol': ISOCodeData('', '', 'Mongo', 'mongo'),
    'loz': ISOCodeData('', '', 'Lozi', 'lozi'),
    'ltz': ISOCodeData(
        '',
        'lb',
        'Luxembourgish; Letzeburgesch',
        'luxembourgeois',
    ),
    'lua': ISOCodeData('', '', 'Luba-Lulua', 'luba-lulua'),
    'lub': ISOCodeData('', 'lu', 'Luba-Katanga', 'luba-katanga'),
    'lug': ISOCodeData('', 'lg', 'Ganda', 'ganda'),
    'lui': ISOCodeData('', '', 'Luiseno', 'luiseno'),
    'lun': ISOCodeData('', '', 'Lunda', 'lunda'),
    'luo': ISOCodeData(
        '',
        '',
        'Luo (Kenya and Tanzania)',
        'luo (Kenya et Tanzanie)',
    ),
    'lus': ISOCodeData('', '', 'Lushai', 'lushai'),
    'mac': ISOCodeData('mkd', 'mk', 'Macedonian', 'macédonien'),
    'mad': ISOCodeData('', '', 'Madurese', 'madourais'),
    'mag': ISOCodeData('', '', 'Magahi', 'magahi'),
    'mah': ISOCodeData('', 'mh', 'Marshallese', 'marshall'),
    'mai': ISOCodeData('', '', 'Maithili', 'maithili'),
    'mak': ISOCodeData('', '', 'Makasar', 'makassar'),
    'mal': ISOCodeData('', 'ml', 'Malayalam', 'malayalam'),
    'man': ISOCodeData('', '', 'Mandingo', 'mandingue'),
    'mao': ISOCodeData('mri', 'mi', 'Maori', 'maori'),
    'map': ISOCodeData(
        '',
        '',
        'Austronesian languages',
        'austronésiennes, langues',
    ),
    'mar': ISOCodeData('', 'mr', 'Marathi', 'marathe'),
    'mas': ISOCodeData('', '', 'Masai', 'massaï'),
    'may': ISOCodeData('msa', 'ms', 'Malay', 'malais'),
    'mdf': ISOCodeData('', '', 'Moksha', 'moksa'),
    'mdr': ISOCodeData('', '', 'Mandar', 'mandar'),
    'men': ISOCodeData('', '', 'Mende', 'mendé'),
    'mga': ISOCodeData(
        '',
        '',
        'Irish, Middle (900-1200)',
        'irlandais moyen (900-1200)',
    ),
    'mic': ISOCodeData('', '', "Mi'kmaq; Micmac", "mi'kmaq; micmac"),
    'min': ISOCodeData('', '', 'Minangkabau', 'minangkabau'),
    'mis': ISOCodeData('', '', 'Uncoded languages', 'langues non codées'),
    'mkh': ISOCodeData('', '', 'Mon-Khmer languages', 'môn-khmer, langues'),
    'mlg': ISOCodeData('', 'mg', 'Malagasy', 'malgache'),
    'mlt': ISOCodeData('', 'mt', 'Maltese', 'maltais'),
    'mnc': ISOCodeData('', '', 'Manchu', 'mandchou'),
    'mni': ISOCodeData('', '', 'Manipuri', 'manipuri'),
    'mno': ISOCodeData('', '', 'Manobo languages', 'manobo, langues'),
    'moh': ISOCodeData('', '', 'Mohawk', 'mohawk'),
    'mon': ISOCodeData('', 'mn', 'Mongolian', 'mongol'),
    'mos': ISOCodeData('', '', 'Mossi', 'moré'),
    'mul': ISOCodeData('', '', 'Multiple languages', 'multilingue'),
    'mun': ISOCodeData('', '', 'Munda languages', 'mounda, langues'),
    'mus': ISOCodeData('', '', 'Creek', 'muskogee'),
    'mwl': ISOCodeData('', '', 'Mirandese', 'mirandais'),
    'mwr': ISOCodeData('', '', 'Marwari', 'marvari'),
    'myn': ISOCodeData('', '', 'Mayan languages', 'maya, langues'),
    'myv': ISOCodeData('', '', 'Erzya', 'erza'),
    'nah': ISOCodeData('', '', 'Nahuatl languages', 'nahuatl, langues'),
    'nai': ISOCodeData(
        '',
        '',
        'North American Indian languages',
        'nord-amérindiennes, langues',
    ),
    'nap': ISOCodeData('', '', 'Neapolitan', 'napolitain'),
    'nau': ISOCodeData('', 'na', 'Nauru', 'nauruan'),
    'nav': ISOCodeData('', 'nv', 'Navajo; Navaho', 'navaho'),
    'nbl': ISOCodeData(
        '',
        'nr',
        'Ndebele, South; South Ndebele',
        'ndébélé du Sud',
    ),
    'nde': ISOCodeData(
        '',
        'nd',
        'Ndebele, North; North Ndebele',
        'ndébélé du Nord',
    ),
    'ndo': ISOCodeData('', 'ng', 'Ndonga', 'ndonga'),
    'nds': ISOCodeData(
        '',
        '',
        'Low German; Low Saxon; German, Low; Saxon, Low',
        'bas allemand; bas saxon; allemand, bas; saxon, bas',
    ),
    'nep': ISOCodeData('', 'ne', 'Nepali', 'népalais'),
    'new': ISOCodeData('', '', 'Nepal Bhasa; Newari', 'nepal bhasa; newari'),
    'nia': ISOCodeData('', '', 'Nias', 'nias'),
    'nic': ISOCodeData(
        '',
        '',
        'Niger-Kordofanian languages',
        'nigéro-kordofaniennes, langues',
    ),
    'niu': ISOCodeData('', '', 'Niuean', 'niué'),
    'nno': ISOCodeData(
        '',
        'nn',
        'Norwegian Nynorsk; Nynorsk, Norwegian',
        'norvégien nynorsk; nynorsk, norvégien',
    ),
    'nob': ISOCodeData(
        '',
        'nb',
        'Bokmål, Norwegian; Norwegian Bokmål',
        'norvégien bokmål',
    ),
    'nog': ISOCodeData('', '', 'Nogai', 'nogaï; nogay'),
    'non': ISOCodeData('', '', 'Norse, Old', 'norrois, vieux'),
    'nor': ISOCodeData('', 'no', 'Norwegian', 'norvégien'),
    'nqo': ISOCodeData('', '', "N'Ko", "n'ko"),
    'nso': ISOCodeData(
        '',
        '',
        'Pedi; Sepedi; Northern Sotho',
        'pedi; sepedi; sotho du Nord',
    ),
    'nub': ISOCodeData('', '', 'Nubian languages', 'nubiennes, langues'),
    'nwc': ISOCodeData(
        '',
        '',
        'Classical Newari; Old Newari; Classical Nepal Bhasa',
        'newari classique',
    ),
    'nya': ISOCodeData(
        '',
        'ny',
        'Chichewa; Chewa; Nyanja',
        'chichewa; chewa; nyanja',
    ),
    'nym': ISOCodeData('', '', 'Nyamwezi', 'nyamwezi'),
    'nyn': ISOCodeData('', '', 'Nyankole', 'nyankolé'),
    'nyo': ISOCodeData('', '', 'Nyoro', 'nyoro'),
    'nzi': ISOCodeData('', '', 'Nzima', 'nzema'),
    'oci': ISOCodeData(
        '',
        'oc',
        'Occitan (post 1500)',
        'occitan (après 1500)',
    ),
    'oji': ISOCodeData('', 'oj', 'Ojibwa', 'ojibwa'),
    'ori': ISOCodeData('', 'or', 'Oriya', 'oriya'),
    'orm': ISOCodeData('', 'om', 'Oromo', 'galla'),
    'osa': ISOCodeData('', '', 'Osage', 'osage'),
    'oss': ISOCodeData('', 'os', 'Ossetian; Ossetic', 'ossète'),
    'ota': ISOCodeData(
        '',
        '',
        'Turkish, Ottoman (1500-1928)',
        'turc ottoman (1500-1928)',
    ),
    'oto': ISOCodeData('', '', 'Otomian languages', 'otomi, langues'),
    'paa': ISOCodeData('', '', 'Papuan languages', 'papoues, langues'),
    'pag': ISOCodeData('', '', 'Pangasinan', 'pangasinan'),
    'pal': ISOCodeData('', '', 'Pahlavi', 'pahlavi'),
    'pam': ISOCodeData('', '', 'Pampanga; Kapampangan', 'pampangan'),
    'pan': ISOCodeData('', 'pa', 'Panjabi; Punjabi', 'pendjabi'),
    'pap': ISOCodeData('', '', 'Papiamento', 'papiamento'),
    'pau': ISOCodeData('', '', 'Palauan', 'palau'),
    'peo': ISOCodeData(
        '',
        '',
        'Persian, Old (ca.600-400 B.C.)',
        'perse, vieux (ca. 600-400 av. J.-C.)',
    ),
    'per': ISOCodeData('fas', 'fa', 'Persian', 'persan'),
    'phi': ISOCodeData(
        '',
        '',
        'Philippine languages',
        'philippines, langues',
    ),
    'phn': ISOCodeData('', '', 'Phoenician', 'phénicien'),
    'pli': ISOCodeData('', 'pi', 'Pali', 'pali'),
    'pol': ISOCodeData('', 'pl', 'Polish', 'polonais'),
    'pon': ISOCodeData('', '', 'Pohnpeian', 'pohnpei'),
    'por': ISOCodeData('', 'pt', 'Portuguese', 'portugais'),
    'pra': ISOCodeData('', '', 'Prakrit languages', 'prâkrit, langues'),
    'pro': ISOCodeData(
        '',
        '',
        'Provençal, Old (to 1500); Occitan, Old (to 1500)',
        "provençal ancien (jusqu'à 1500); occitan ancien (jusqu'à 1500)",
    ),
    'pus': ISOCodeData('', 'ps', 'Pushto; Pashto', 'pachto'),
    'qaa': ISOCodeData(
        '',
        '',
        'Reserved for local use',
        "réservée à l'usage local",
    ),
    'que': ISOCodeData('', 'qu', 'Quechua', 'quechua'),
    'raj': ISOCodeData('', '', 'Rajasthani', 'rajasthani'),
    'rap': ISOCodeData('', '', 'Rapanui', 'rapanui'),
    'rar': ISOCodeData(
        '',
        '',
        'Rarotongan; Cook Islands Maori',
        'rarotonga; maori des îles Cook',
    ),
    'roa': ISOCodeData('', '', 'Romance languages', 'romanes, langues'),
    'roh': ISOCodeData('', 'rm', 'Romansh', 'romanche'),
    'rom': ISOCodeData('', '', 'Romany', 'tsigane'),
    'rum': ISOCodeData(
        'ron',
        'ro',
        'Romanian; Moldavian; Moldovan',
        'roumain; moldave',
    ),
    'run': ISOCodeData('', 'rn', 'Rundi', 'rundi'),
    'rup': ISOCodeData(
        '',
        '',
        'Aromanian; Arumanian; Macedo-Romanian',
        'aroumain; macédo-roumain',
    ),
    'rus': ISOCodeData('', 'ru', 'Russian', 'russe'),
    'sad': ISOCodeData('', '', 'Sandawe', 'sandawe'),
    'sag': ISOCodeData('', 'sg', 'Sango', 'sango'),
    'sah': ISOCodeData('', '', 'Yakut', 'iakoute'),
    'sai': ISOCodeData(
        '',
        '',
        'South American Indian languages',
        'sud-amérindiennes, langues',
    ),
    'sal': ISOCodeData('', '', 'Salishan languages', 'salishennes, langues'),
    'sam': ISOCodeData('', '', 'Samaritan Aramaic', 'samaritain'),
    'san': ISOCodeData('', 'sa', 'Sanskrit', 'sanskrit'),
    'sas': ISOCodeData('', '', 'Sasak', 'sasak'),
    'sat': ISOCodeData('', '', 'Santali', 'santal'),
    'scn': ISOCodeData('', '', 'Sicilian', 'sicilien'),
    'sco': ISOCodeData('', '', 'Scots', 'écossais'),
    'sel': ISOCodeData('', '', 'Selkup', 'selkoupe'),
    'sem': ISOCodeData('', '', 'Semitic languages', 'sémitiques, langues'),
    'sga': ISOCodeData(
        '',
        '',
        'Irish, Old (to 900)',
        "irlandais ancien (jusqu'à 900)",
    ),
    'sgn': ISOCodeData('', '', 'Sign Languages', 'langues des signes'),
    'shn': ISOCodeData('', '', 'Shan', 'chan'),
    'sid': ISOCodeData('', '', 'Sidamo', 'sidamo'),
    'sin': ISOCodeData('', 'si', 'Sinhala; Sinhalese', 'singhalais'),
    'sio': ISOCodeData('', '', 'Siouan languages', 'sioux, langues'),
    'sit': ISOCodeData(
        '',
        '',
        'Sino-Tibetan languages',
        'sino-tibétaines, langues',
    ),
    'sla': ISOCodeData('', '', 'Slavic languages', 'slaves, langues'),
    'slo': ISOCodeData('slk', 'sk', 'Slovak', 'slovaque'),
    'slv': ISOCodeData('', 'sl', 'Slovenian', 'slovène'),
    'sma': ISOCodeData('', '', 'Southern Sami', 'sami du Sud'),
    'sme': ISOCodeData('', 'se', 'Northern Sami', 'sami du Nord'),
    'smi': ISOCodeData('', '', 'Sami languages', 'sames, langues'),
    'smj': ISOCodeData('', '', 'Lule Sami', 'sami de Lule'),
    'smn': ISOCodeData('', '', 'Inari Sami', "sami d'Inari"),
    'smo': ISOCodeData('', 'sm', 'Samoan', 'samoan'),
    'sms': ISOCodeData('', '', 'Skolt Sami', 'sami skolt'),
    'sna': ISOCodeData('', 'sn', 'Shona', 'shona'),
    'snd': ISOCodeData('', 'sd', 'Sindhi', 'sindhi'),
    'snk': ISOCodeData('', '', 'Soninke', 'soninké'),
    'sog': ISOCodeData('', '', 'Sogdian', 'sogdien'),
    'som': ISOCodeData('', 'so', 'Somali', 'somali'),
    'son': ISOCodeData('', '', 'Songhai languages', 'songhai, langues'),
    'sot': ISOCodeData('', 'st', 'Sotho, Southern', 'sotho du Sud'),
    'spa': ISOCodeData('', 'es', 'Spanish; Castilian', 'espagnol; castillan'),
    'srd': ISOCodeData('', 'sc', 'Sardinian', 'sarde'),
    'srn': ISOCodeData('', '', 'Sranan Tongo', 'sranan tongo'),
    'srp': ISOCodeData('', 'sr', 'Serbian', 'serbe'),
    'srr': ISOCodeData('', '', 'Serer', 'sérère'),
    'ssa': ISOCodeData(
        '',
        '',
        'Nilo-Saharan languages',
        'nilo-sahariennes, langues',
    ),
    'ssw': ISOCodeData('', 'ss', 'Swati', 'swati'),
    'suk': ISOCodeData('', '', 'Sukuma', 'sukuma'),
    'sun': ISOCodeData('', 'su', 'Sundanese', 'soundanais'),
    'sus': ISOCodeData('', '', 'Susu', 'soussou'),
    'sux': ISOCodeData('', '', 'Sumerian', 'sumérien'),
    'swa': ISOCodeData('', 'sw', 'Swahili', 'swahili'),
    'swe': ISOCodeData('', 'sv', 'Swedish', 'suédois'),
    'syc': ISOCodeData('', '', 'Classical Syriac', 'syriaque classique'),
    'syr': ISOCodeData('', '', 'Syriac', 'syriaque'),
    'tah': ISOCodeData('', 'ty', 'Tahitian', 'tahitien'),
    'tai': ISOCodeData('', '', 'Tai languages', 'tai, langues'),
    'tam': ISOCodeData('', 'ta', 'Tamil', 'tamoul'),
    'tat': ISOCodeData('', 'tt', 'Tatar', 'tatar'),
    'tel': ISOCodeData('', 'te', 'Telugu', 'télougou'),
    'tem': ISOCodeData('', '', 'Timne', 'temne'),
    'ter': ISOCodeData('', '', 'Tereno', 'tereno'),
    'tet': ISOCodeData('', '', 'Tetum', 'tetum'),
    'tgk': ISOCodeData('', 'tg', 'Tajik', 'tadjik'),
    'tgl': ISOCodeData('', 'tl', 'Tagalog', 'tagalog'),
    'tha': ISOCodeData('', 'th', 'Thai', 'thaï'),
    'tib': ISOCodeData('bod', 'bo', 'Tibetan', 'tibétain'),
    'tig': ISOCodeData('', '', 'Tigre', 'tigré'),
    'tir': ISOCodeData('', 'ti', 'Tigrinya', 'tigrigna'),
    'tiv': ISOCodeData('', '', 'Tiv', 'tiv'),
    'tkl': ISOCodeData('', '', 'Tokelau', 'tokelau'),
    'tlh': ISOCodeData('', '', 'Klingon; tlhIngan-Hol', 'klingon'),
    'tli': ISOCodeData('', '', 'Tlingit', 'tlingit'),
    'tmh': ISOCodeData('', '', 'Tamashek', 'tamacheq'),
    'tog': ISOCodeData('', '', 'Tonga (Nyasa)', 'tonga (Nyasa)'),
    'ton': ISOCodeData(
        '',
        'to',
        'Tonga (Tonga Islands)',
        'tongan (Îles Tonga)',
    ),
    'tpi': ISOCodeData('', '', 'Tok Pisin', 'tok pisin'),
    'tsi': ISOCodeData('', '', 'Tsimshian', 'tsimshian'),
    'tsn': ISOCodeData('', 'tn', 'Tswana', 'tswana'),
    'tso': ISOCodeData('', 'ts', 'Tsonga', 'tsonga'),
    'tuk': ISOCodeData('', 'tk', 'Turkmen', 'turkmène'),
    'tum': ISOCodeData('', '', 'Tumbuka', 'tumbuka'),
    'tup': ISOCodeData('', '', 'Tupi languages', 'tupi, langues'),
    'tur': ISOCodeData('', 'tr', 'Turkish', 'turc'),
    'tut': ISOCodeData('', '', 'Altaic languages', 'altaïques, langues'),
    'tvl': ISOCodeData('', '', 'Tuvalu', 'tuvalu'),
    'twi': ISOCodeData('', 'tw', 'Twi', 'twi'),
    'tyv': ISOCodeData('', '', 'Tuvinian', 'touva'),
    'udm': ISOCodeData('', '', 'Udmurt', 'oudmourte'),
    'uga': ISOCodeData('', '', 'Ugaritic', 'ougaritique'),
    'uig': ISOCodeData('', 'ug', 'Uighur; Uyghur', 'ouïgour'),
    'ukr': ISOCodeData('', 'uk', 'Ukrainian', 'ukrainien'),
    'umb': ISOCodeData('', '', 'Umbundu', 'umbundu'),
    'und': ISOCodeData('', '', 'Undetermined', 'indéterminée'),
    'urd': ISOCodeData('', 'ur', 'Urdu', 'ourdou'),
    'uzb': ISOCodeData('', 'uz', 'Uzbek', 'ouszbek'),
    'vai': ISOCodeData('', '', 'Vai', 'vaï'),
    'ven': ISOCodeData('', 've', 'Venda', 'venda'),
    'vie': ISOCodeData('', 'vi', 'Vietnamese', 'vietnamien'),
    'vol': ISOCodeData('', 'vo', 'Volapük', 'volapük'),
    'vot': ISOCodeData('', '', 'Votic', 'vote'),
    'wak': ISOCodeData('', '', 'Wakashan languages', 'wakashanes, langues'),
    'wal': ISOCodeData('', '', 'Wolaitta; Wolaytta', 'wolaitta; wolaytta'),
    'war': ISOCodeData('', '', 'Waray', 'waray'),
    'was': ISOCodeData('', '', 'Washo', 'washo'),
    'wel': ISOCodeData('cym', 'cy', 'Welsh', 'gallois'),
    'wen': ISOCodeData('', '', 'Sorbian languages', 'sorabes, langues'),
    'wln': ISOCodeData('', 'wa', 'Walloon', 'wallon'),
    'wol': ISOCodeData('', 'wo', 'Wolof', 'wolof'),
    'xal': ISOCodeData('', '', 'Kalmyk; Oirat', 'kalmouk; oïrat'),
    'xho': ISOCodeData('', 'xh', 'Xhosa', 'xhosa'),
    'yao': ISOCodeData('', '', 'Yao', 'yao'),
    'yap': ISOCodeData('', '', 'Yapese', 'yapois'),
    'yid': ISOCodeData('', 'yi', 'Yiddish', 'yiddish'),
    'yor': ISOCodeData('', 'yo', 'Yoruba', 'yoruba'),
    'ypk': ISOCodeData('', '', 'Yupik languages', 'yupik, langues'),
    'zap': ISOCodeData('', '', 'Zapotec', 'zapotèque'),
    'zbl': ISOCodeData(
        '',
        '',
        'Blissymbols; Blissymbolics; Bliss',
        'symboles Bliss; Bliss',
    ),
    'zen': ISOCodeData('', '', 'Zenaga', 'zenaga'),
    'zgh': ISOCodeData(
        '',
        '',
        'Standard Moroccan Tamazight',
        'amazighe standard marocain',
    ),
    'zha': ISOCodeData('', 'za', 'Zhuang; Chuang', 'zhuang; chuang'),
    'znd': ISOCodeData('', '', 'Zande languages', 'zandé, langues'),
    'zul': ISOCodeData('', 'zu', 'Zulu', 'zoulou'),
    'zun': ISOCodeData('', '', 'Zuni', 'zuni'),
    'zxx': ISOCodeData(
        '',
        '',
        'No linguistic content; Not applicable',
        'pas de contenu linguistique; non applicable',
    ),
    'zza': ISOCodeData(
        '',
        '',
        'Zaza; Dimili; Dimli; Kirdki; Kirmanjki; Zazaki',
        'zaza; dimili; dimli; kirdki; kirmanjki; zazaki',
    ),
}


def iso_639_2_from_3(iso3: str) -> str:
    """Convert ISO 639-3 code to ISO 639-2 code."""
    if iso3 in ISO_639_3:
        return ISO_639_3[iso3].alpha_2
    else:
        return ""



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

from io import BytesIO
from os import fspath
from pathlib import Path
from unittest.mock import MagicMock, patch

import img2pdf
import pikepdf
import pytest
from pikepdf import Array, Dictionary, Name
from PIL import Image, ImageDraw

from ocrmypdf import optimize as opt
from ocrmypdf._exec import jbig2enc, pngquant
from ocrmypdf._exec.ghostscript import rasterize_pdf
from ocrmypdf.helpers import IMG2PDF_KWARGS, Resolution
from ocrmypdf.optimize import PdfImage, extract_image_filter
from tests.conftest import check_ocrmypdf

needs_pngquant = pytest.mark.skipif(
    not pngquant.available(), reason="pngquant not installed"
)
needs_jbig2enc = pytest.mark.skipif(
    not jbig2enc.available(), reason="jbig2enc not installed"
)


# pylint:disable=redefined-outer-name


@pytest.fixture(scope="session")
def palette(resources):
    return resources / 'palette.pdf'


@needs_pngquant
@pytest.mark.parametrize('pdf', ['multipage', 'palette'])
def test_basic(multipage, palette, pdf, outpdf):
    infile = multipage if pdf == 'multipage' else palette
    opt.main(infile, outpdf, level=3)

    assert 0.98 * Path(outpdf).stat().st_size <= Path(infile).stat().st_size


@needs_pngquant
def test_mono_not_inverted(resources, outdir):
    infile = resources / '2400dpi.pdf'
    opt.main(infile, outdir / 'out.pdf', level=3)

    rasterize_pdf(
        outdir / 'out.pdf',
        outdir / 'im.png',
        raster_device='pnggray',
        raster_dpi=Resolution(10, 10),
    )

    with Image.open(fspath(outdir / 'im.png')) as im:
        assert im.getpixel((0, 0)) > 240, "Expected white background"


@needs_pngquant
def test_jpg_png_params(resources, outpdf):
    check_ocrmypdf(
        resources / 'crom.png',
        outpdf,
        '--image-dpi',
        '200',
        '--optimize',
        '3',
        '--jpg-quality',
        '50',
        '--png-quality',
        '20',
        '--plugin',
        'tests/plugins/tesseract_noop.py',
    )


@needs_jbig2enc
@pytest.mark.parametrize('lossy', [False, True])
def test_jbig2_lossy(lossy, resources, outpdf):
    args = [
        resources / 'ccitt.pdf',
        outpdf,
        '--image-dpi',
        '200',
        '--optimize',
        '3',
        '--jpg-quality',
        '50',
        '--png-quality',
        '20',
        '--plugin',
        'tests/plugins/tesseract_noop.py',
        '--jbig2-threshold',
        '0.7',
    ]
    if lossy:
        args.append('--jbig2-lossy')

    check_ocrmypdf(*args)

    with pikepdf.open(outpdf) as pdf:
        pim = pikepdf.PdfImage(next(iter(pdf.pages[0].images.values())))
        assert pim.filters[0] == '/JBIG2Decode'

        if lossy:
            assert '/JBIG2Globals' in pim.decode_parms[0]
        else:
            assert len(pim.decode_parms) == 0


@needs_pngquant
@needs_jbig2enc
def test_flate_to_jbig2(resources, outdir):
    # This test requires an image that pngquant is capable of converting to
    # to 1bpp - so use an existing 1bpp image, convert up, confirm it can
    # convert down
    with Image.open(fspath(resources / 'typewriter.png')) as im:
        assert im.mode in ('1', 'P')
        im = im.convert('L')
        im.save(fspath(outdir / 'type8.png'))

    check_ocrmypdf(
        outdir / 'type8.png',
        outdir / 'out.pdf',
        '--image-dpi',
        '100',
        '--png-quality',
        '50',
        '--optimize',
        '3',
        '--plugin',
        'tests/plugins/tesseract_noop.py',
    )

    with pikepdf.open(outdir / 'out.pdf') as pdf:
        pim = pikepdf.PdfImage(next(iter(pdf.pages[0].images.values())))
        assert pim.filters[0] == '/JBIG2Decode'


@needs_pngquant
def test_multiple_pngs(resources, outdir):
    with Path.open(outdir / 'in.pdf', 'wb') as inpdf:
        img2pdf.convert(
            fspath(resources / 'baiona_colormapped.png'),
            fspath(resources / 'baiona_gray.png'),
            outputstream=inpdf,
            **IMG2PDF_KWARGS,
        )

    def mockquant(input_file, output_file, *_args):
        with Image.open(input_file) as im:
            draw = ImageDraw.Draw(im)
            draw.rectangle((0, 0, im.width, im.height), fill=128)
            im.save(output_file)

    with patch('ocrmypdf.optimize.pngquant.quantize') as mock:
        mock.side_effect = mockquant
        check_ocrmypdf(
            outdir / 'in.pdf',
            outdir / 'out.pdf',
            '--optimize',
            '3',
            '--jobs',
            '1',
            '--use-threads',
            '--output-type',
            'pdf',
            '--plugin',
            'tests/plugins/tesseract_noop.py',
        )
        mock.assert_called()

    with (
        pikepdf.open(outdir / 'in.pdf') as inpdf,
        pikepdf.open(outdir / 'out.pdf') as outpdf,
    ):
        for n in range(len(inpdf.pages)):
            inim = next(iter(inpdf.pages[n].images.values()))
            outim = next(iter(outpdf.pages[n].images.values()))
            assert len(outim.read_raw_bytes()) < len(inim.read_raw_bytes()), n


def test_optimize_off(resources, outpdf):
    check_ocrmypdf(
        resources / 'trivial.pdf',
        outpdf,
        '--optimize=0',
        '--output-type',
        'pdf',
        '--plugin',
        'tests/plugins/tesseract_noop.py',
    )


def test_group3(resources):
    with pikepdf.open(resources / 'ccitt.pdf') as pdf:
        im = pdf.pages[0].Resources.XObject['/Im1']
        assert (
            opt.extract_image_filter(im, im.objgen[0]) is not None
        ), "Group 4 should be allowed"

        im.DecodeParms['/K'] = 0
        assert (
            opt.extract_image_filter(im, im.objgen[0]) is None
        ), "Group 3 should be disallowed"


def test_find_formx(resources):
    with pikepdf.open(resources / 'formxobject.pdf') as pdf:
        working, pagenos = opt._find_image_xrefs(pdf)
        assert len(working) == 1
        xref = next(iter(working))
        assert pagenos[xref] == 0


def test_extract_image_filter_with_pdf_image():
    image = MagicMock()
    image.Subtype = Name.Image
    image.Length = 200
    image.Width = 10
    image.Height = 10
    image.Filter = [Name.FlateDecode, Name.DCTDecode]
    pdf_image = PdfImage(image)
    image.BitsPerComponent = 8
    assert extract_image_filter(image, None) == (
        pdf_image,
        pdf_image.filter_decodeparms[1],
    )


def test_extract_image_filter_with_non_image():
    image = MagicMock()
    image.Subtype = Name.Form
    assert extract_image_filter(image, None) is None


def test_extract_image_filter_with_small_stream_size():
    image = MagicMock()
    image.Subtype = Name.Image
    image.Length = 50
    assert extract_image_filter(image, None) is None


def test_extract_image_filter_with_small_dimensions():
    image = MagicMock()
    image.Subtype = Name.Image
    image.Length = 200
    image.Width = 5
    image.Height = 5
    assert extract_image_filter(image, None) is None


def test_extract_image_filter_with_multiple_compression_filters():
    image = MagicMock()
    image.Subtype = Name.Image
    image.Length = 200
    image.Width = 10
    image.Height = 10
    image.BitsPerComponent = 8
    image.Filter = [Name.ASCII85Decode, Name.FlateDecode, Name.DCTDecode]
    assert extract_image_filter(image, None) is None


def test_extract_image_filter_with_wide_gamut_image():
    image = MagicMock()
    image.Subtype = Name.Image
    image.Length = 200
    image.Width = 10
    image.Height = 10
    image.BitsPerComponent = 16
    image.Filter = Name.FlateDecode
    assert extract_image_filter(image, None) is None


def test_extract_image_filter_with_jpeg2000_image():
    im = Image.new('RGB', (10, 10))
    bio = BytesIO()
    im.save(bio, format='JPEG2000')
    pdf = pikepdf.new()
    stream = pdf.make_stream(
        data=bio.getvalue(),
        Subtype=Name.Image,
        Length=200,
        Width=10,
        Height=10,
        BitsPerComponent=8,
        Filter=Name.JPXDecode,
    )
    assert extract_image_filter(stream, None) is None


def test_extract_image_filter_with_ccitt_group_3_image():
    image = MagicMock()
    image.Subtype = Name.Image
    image.Length = 200
    image.Width = 10
    image.Height = 10
    image.BitsPerComponent = 1
    image.Filter = Name.CCITTFaxDecode
    image.DecodeParms = Array([Dictionary(K=1)])
    assert extract_image_filter(image, None) is None


# Triggers pikepdf bug
# def test_extract_image_filter_with_decode_table():
#     image = MagicMock()
#     image.Subtype = Name.Image
#     image.Length = 200
#     image.Width = 10
#     image.Height = 10
#     image.Filter = Name.FlateDecode
#     image.BitsPerComponent = 8
#     image.ColorSpace = Name.DeviceGray
#     image.Decode = [42, 0]
#     assert extract_image_filter(image, None) is None



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""Utilities for PDF/A production and confirmation with Ghostspcript."""

from __future__ import annotations

import base64
from collections.abc import Iterator
from importlib.resources import files as package_files
from pathlib import Path

import pikepdf

SRGB_ICC_PROFILE_NAME = 'sRGB.icc'


def _postscript_objdef(
    alias: str,
    dictionary: dict[str, str],
    *,
    stream_name: str | None = None,
    stream_data: bytes | None = None,
) -> Iterator[str]:
    assert (stream_name is None) == (stream_data is None)

    objtype = '/stream' if stream_name else '/dict'

    if stream_name:
        assert stream_data is not None
        a85_data = base64.a85encode(stream_data, adobe=True).decode('ascii')
        yield f'{stream_name} ' + a85_data
        yield 'def'

    if alias != '{Catalog}':  # Catalog needs no definition
        yield f'[/_objdef {alias} /type {objtype} /OBJ pdfmark'

    yield f'[{alias} <<'
    for key, val in dictionary.items():
        yield f'  {key} {val}'
    yield '>> /PUT pdfmark'

    if stream_name:
        yield f'[{alias} {stream_name[1:]} /PUT pdfmark'


def _make_postscript(icc_name: str, icc_data: bytes, colors: int) -> Iterator[str]:
    yield '%!'
    yield from _postscript_objdef(
        '{icc_PDFA}',  # Not an f-string
        {'/N': str(colors)},
        stream_name='/ICCProfile',
        stream_data=icc_data,
    )
    yield ''
    yield from _postscript_objdef(
        '{OutputIntent_PDFA}',
        {
            '/Type': '/OutputIntent',
            '/S': '/GTS_PDFA1',
            '/DestOutputProfile': '{icc_PDFA}',
            '/OutputConditionIdentifier': f'({icc_name})',  # Only f-string
        },
    )
    yield ''
    yield from _postscript_objdef(
        '{Catalog}', {'/OutputIntents': '[ {OutputIntent_PDFA} ]'}
    )


def generate_pdfa_ps(target_filename: Path, icc: str = 'sRGB'):
    """Create a Postscript PDFMARK file for Ghostscript PDF/A conversion.

    pdfmark is an extension to the Postscript language that describes some PDF
    features like bookmarks and annotations. It was originally specified Adobe
    Distiller, for Postscript to PDF conversion.

    Ghostscript uses pdfmark for PDF to PDF/A conversion as well. To use Ghostscript
    to create a PDF/A, we need to create a pdfmark file with the necessary metadata.

    This function takes care of the many version-specific bugs and peculiarities in
    Ghostscript's handling of pdfmark.

    The only information we put in specifies that we want the file to be a
    PDF/A, and we want to Ghostscript to convert objects to the sRGB colorspace
    if it runs into any object that it decides must be converted.

    Arguments:
        target_filename: filename to save
        icc: ICC identifier such as 'sRGB'
    References:
        Adobe PDFMARK Reference:
        https://opensource.adobe.com/dc-acrobat-sdk-docs/library/pdfmark/
    """
    if icc != 'sRGB':
        raise NotImplementedError("Only supporting sRGB")

    bytes_icc_profile = (
        package_files('ocrmypdf.data') / SRGB_ICC_PROFILE_NAME
    ).read_bytes()
    postscript = '\n'.join(_make_postscript(icc, bytes_icc_profile, 3))

    # We should have encoded everything to pure ASCII by this point, and
    # to be safe, only allow ASCII in PostScript
    Path(target_filename).write_text(postscript, encoding='ascii')
    return target_filename


def file_claims_pdfa(filename: Path):
    """Determines if the file claims to be PDF/A compliant.

    This only checks if the XMP metadata contains a PDF/A marker. It does not
    do full PDF/A validation.
    """
    with pikepdf.open(filename) as pdf:
        pdfmeta = pdf.open_metadata()
        if not pdfmeta.pdfa_status:
            return {
                'pass': False,
                'output': 'pdf',
                'conformance': 'No PDF/A metadata in XMP',
            }
        valid_part_conforms = {'1A', '1B', '2A', '2B', '2U', '3A', '3B', '3U'}
        conformance = f'PDF/A-{pdfmeta.pdfa_status}'
        pdfa_dict: dict[str, str | bool] = {}
        if pdfmeta.pdfa_status in valid_part_conforms:
            pdfa_dict['pass'] = True
            pdfa_dict['output'] = 'pdfa'
        pdfa_dict['conformance'] = conformance
    return pdfa_dict
